#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the class definition for the MacroServer environment
manager"""

__all__ = ["EnvironmentManager"]

__docformat__ = 'restructuredtext'

import os
import shelve
import operator

from taurus.core.util import CaselessDict

from msmanager import MacroServerManager
from msexception import UnknownEnv


class EnvironmentManager(MacroServerManager):
    """The MacroServer environment manager class. It is designed to be a
    singleton for the entire application.
    """

    def __init__(self, macro_server, environment_db=None):
        MacroServerManager.__init__(self, macro_server)
        if environment_db is not None:
            self.setEnvironmentDb(environment_db)

    def reInit(self):
        """(Re)initializes the manager"""
        if self.is_initialized():
            return

        # a string containing the absolute filename containing the environment
        self._env_name = None

        # the full enviroment (a shelf for now - can be accessed as a dict)
        self._env = None

        # cache environment for keys that start with door name
        # dict<string, dict<string, value> > where:
        #  - key: door name
        #  - value: dict where:
        #    - key: environment name
        #    - value: environment value
        self._door_env = None

        # cache environment for keys that start with macro name
        # dict<string, dict<string, value> > where:
        #  - key: macro name
        #  - value: dict where:
        #    - key: environment name
        #    - value: environment value
        self._macro_env = None

        # cache environment for global keys
        # dict<string, value> where:
        #  - key: environment name
        #  - value: environment value
        self._global_env = None

        self._initEnv()

        MacroServerManager.reInit(self)

    def cleanUp(self):
        if self.is_cleaned():
            return

        self._clearEnv()

        MacroServerManager.cleanUp(self)

    def _initEnv(self):
        self._macro_env, self._global_env = {}, {}
        self._door_env = CaselessDict()

    def _clearEnv(self):
        self._env = self._macro_env = self._global_env = self._door_env = None

    def setEnvironmentDb(self, f_name):
        """Sets up a new environment from a file"""
        self._initEnv()
        f_name = os.path.abspath(f_name)
        self._env_name = f_name
        dir_name = os.path.dirname(f_name)
        if not os.path.isdir(dir_name):
            try:
                self.info("Creating environment directory: %s" % dir_name)
                os.makedirs(dir_name)
            except OSError, ose:
                self.error("Creating environment: %s" % ose.strerror)
                self.debug("Details:", exc_info=1)
                raise ose
        try:
            self._env = shelve.open(f_name, flag='c', protocol=0,
                                    writeback=False)
        except:
            self.error("Failed to create/access environment in %s", f_name)
            self.debug("Details:", exc_info=1)
            raise

        self.info("Environment is being stored in %s", f_name)

        # fill the three environment caches
        try:
            self._fillEnvironmentCaches(self._env)
        except:
            self.error("Failed to fill local enviroment cache")
            self.debug("Details:", exc_info=1)

    def _fillEnvironmentCaches(self, env):
        # fill the three environment caches
        env_dict = self._global_env
        for k, v in env.items():
            k_parts = k.split('.', 1)
            key = k_parts[0]

            # door or macro property
            if len(k_parts) == 2:
                obj_name, simple_key_name = k_parts
                if obj_name.count('/') == 2:
                    class_dict = self._door_env
                else:
                    class_dict = self._macro_env
                obj_dict = class_dict.get(obj_name, None)
                if obj_dict is None:
                    class_dict[obj_name] = obj_dict = {}
                obj_dict[simple_key_name] = v
            else:
                env_dict[key] = v

    def hasEnv(self, key, macro_name=None, door_name=None):
        #<door>.<macro>.<property name> (highest priority)
        if macro_name and door_name:
            has = self._hasDoorMacroPropertyEnv((door_name, macro_name, key))
            if has:
                return True

        # <macro>.<property name>
        if macro_name:
            has = self._hasMacroPropertyEnv((macro_name, key))
            if has:
                return True

        # <door>.<property name>
        if door_name:
            has = self._hasDoorPropertyEnv((door_name, key))
            if has:
                return True

        # <property name> (less priority)
        return self._hasEnv(key)

    def _getDoorMacroPropertyEnv(self, prop):
        """Returns the property value for a property which must have the
        format <door name>.<macro name>.<property name>"""
        if isinstance(prop, str):
            door_name, macro_name_key = prop.split('.', 1)
        else:
            door_name, macro_name_key = prop[0], '.'.join(prop[1:])
        door_props = self._door_env.get(door_name)
        if door_props is None:
            return None
        return door_props.get(macro_name_key)

    def _hasDoorMacroPropertyEnv(self, prop):
        """Determines if the environment contains a property with the format
        <door name>.<macro name>.<property name>"""
        return not self._getDoorMacroPropertyEnv() is None

    def _getMacroPropertyEnv(self, prop):
        """Returns the property value for a property which must have the
        format <macro name>.<property name>"""
        if isinstance(prop, str):
            macro_name, key = prop.split('.')
        else:
            macro_name, key = prop
        macro_props = self._macro_env.get(macro_name)
        if macro_props is None:
            return None
        return macro_props.get(key)

    def _hasMacroPropertyEnv(self, prop):
        """Determines if the environment contains a property with the format
        <macro name>.<property name>"""
        return not self._getMacroPropertyEnv() is None

    def _getDoorPropertyEnv(self, prop):
        """Returns the property value for a property which must have the
        format <door name>.<property name>"""
        if isinstance(prop, str):
            door_name, key = prop.split('.')
        else:
            door_name, key = prop
        door_props = self._door_env.get(door_name)
        if door_props is None:
            return None
        return door_props.get(key)

    def _hasDoorPropertyEnv(self, prop):
        """Determines if the environment contains a property with the format
        <door name>.<property name>"""
        return not self._getDoorPropertyEnv() is None

    def _getEnv(self, prop):
        """Returns the property value for a property which must have the
        format <property name>"""
        return self._global_env.get(prop)

    def _hasEnv(self, prop):
        """Determines if the environment contains a property with the format
        <property name>"""
        return not self._getEnv(prop) is None

    def getEnv(self, key=None, door_name=None, macro_name=None):
        """Gets the environment matching the given parameters:
        - If key is None it returns the complete environment for the given
          macro and/or door. If both are None the the complete environment is
          returned
          @param[in]"""
        if key is None:
            return self._getAllEnv(door_name=door_name, macro_name=macro_name)

        #<door>.<macro>.<property name> (highest priority)
        if macro_name and door_name:
            v = self._getDoorMacroPropertyEnv((door_name, macro_name, key))
            if not v is None:
                return v

        # <macro>.<property name>
        if macro_name:
            v = self._getMacroPropertyEnv((macro_name, key))
            if not v is None:
                return v

        # <door>.<property name>
        if door_name:
            v = self._getDoorPropertyEnv((door_name, key))
            if not v is None:
                return v

        # <property name> (less priority)
        v = self._getEnv(key)
        if v is None:
            raise UnknownEnv("Unknown environment %s" % key)
        return v

    def _getAllEnv(self, door_name=None, macro_name=None):
        """Gets the complete environment for the given macro and/or door. If
        both are None the the complete environment is returned"""
        if macro_name is None and door_name is None:
            return dict(self._env)
        elif not door_name is None and macro_name is None:
            return self.getDoorEnv(door_name)
        elif door_name and macro_name:
            return self.getAllDoorMacroEnv(door_name, macro_name)
        elif not macro_name is None and door_name is None:
            return self._macro_env.get(macro_name, {})

    def getAllDoorEnv(self, door_name):
        """Gets the complete environment for the given door."""
        door_name = door_name.lower()

        # first go through the global environment
        ret = self._global_env.copy()

        # Then go through the door specific environment
        ret.update(self._door_env.get(door_name, {}))
        return ret

    def getAllDoorMacroEnv(self, door_name, macro_name):
        """Gets the complete environment for the given macro in a specific
        door.

        :param door_name:  the door name (case insensitive)
        :type door_name: str
        :param macro_name: the macro name
        :type macro_name: str

        :return: a dictionary with the resulting environment"""
        door_name = door_name.lower()

        # first go through the global environment
        ret = self._global_env.copy()

        # get the specific door environment
        d_env = self._door_env.get(door_name, {})

        # get the specific macro environment
        m_env = self._macro_env.get(macro_name, {})

        # put the doors global environment
        for k, v in d_env.iteritems():
            if k.count('.') == 0:
                ret[k] = v

        # put the macro environment
        ret.update(m_env)

        # put the door and macro specific environment
        for k, v in d_env.iteritems():
            if k.count('.') > 0:
                m_name, key = k.split('.', 1)
                if m_name is macro_name:
                    ret[key] = v

        return ret

    def getDoorMacroEnv(self, door_name, macro_name, keys=None):
        """Gets the environment for the given macro in a specific door for the
        given key(s)

        :param door_name: the door name (case insensitive)
        :param macro_name: the macro name (case sensitive)
        :param key: the keys to be retrieved. If None (default) the complete
                    environment is returned (same as getAllDoorMacroEnv)
                    key can be a string or a sequence<string>.
                    keys must NOT contain '.' characters

        :return: a dictionary with the resulting environment"""
        if keys is None:
            return self.getAllDoorMacroEnv(door_name, macro_name)

        if isinstance(keys, (str, unicode)):
            keys = (keys,)

        door_name = door_name.lower()

        g_env = self._global_env
        # get the specific door environment
        d_env = self._door_env.get(door_name, {})
        # get the specific macro environment
        m_env = self._macro_env.get(macro_name, {})

        # first go through the global environment
        ret = {}
        for k in keys:
            comp_key = '%s.%s' % (macro_name, k)
            if comp_key in d_env:
                ret[k] = d_env[comp_key]
            elif k in m_env:
                ret[k] = m_env[k]
            elif k in d_env:
                ret[k] = d_env[k]
            elif k in g_env:
                ret[k] = g_env[k]

        return ret

    def _pairwise(self, iterable):
        itnext = iter(iterable).next
        while True:
            yield itnext(), itnext()

    def _dictFromSequence(self, seq):
        return dict(self._pairwise(seq))

    def _encode(self, d):
        ret = {}
        for k, v in d.iteritems():
            if isinstance(v, (str, unicode)):
                try:
                    v = eval(v)
                except:
                    v_lower = v.lower()
                    try:
                        v = eval(v_lower.capitalize())
                    except:
                        pass
            ret[k] = v
        return ret

    def _getCacheForKey(self, key):
        """Returns the cache dictionary object for the given key

        :param key: a string representing the key
        :return: a tuple pair. The first element is the dictionary and the
                 second is the modified key that is applicable to the
                 dictionary"""
        d = None
        key_parts = key.split('.')
        # global property
        if len(key_parts) == 1:
            d = self._global_env
        # macro property
        elif len(key_parts) == 2 and key_parts[0].count('/') != 2:
            macro_name, key = key_parts
            d = self._macro_env.get(macro_name)
            if d is None:
                self._macro_env[macro_name] = d = {}
        # door property
        else:
            door_name, key = key.split('.', 1)
            d = self._door_env.get(door_name)
            if d is None:
                self._door_env[door_name] = d = {}
        return d, key

    def _setOneEnv(self, key, value):
        self._env[key] = value
        self._env.sync()
        d, key = self._getCacheForKey(key)
        d[key] = value

    def _unsetOneEnv(self, key):
        if not key in self._env:
            raise UnknownEnv("Unknown environment %s" % key)
        del self._env[key]
        self._env.sync()
        d, key = self._getCacheForKey(key)
        if key in d:
            del d[key]

    def _unsetEnv(self, env_names):
        for key in env_names:
            self._unsetOneEnv(key)

    def setEnvObj(self, obj):
        """Sets the environment for the given object. If object is a sequence
        then each pair of elements k, v is added as env[k] = v.
        If object is a map then the environmnent is updated.
        Other object types are not supported
        The elements which are strings are 'python evaluated'

        @throws TypeError is obj is not a sequence or a map

        @param[in] obj object to be added to the environment

        @return a dict representing the added environment"""

        if operator.isSequenceType(obj) and \
           not isinstance(obj, (str, unicode)):
            obj = self._dictFromSequence(obj)
        elif not operator.isMappingType(obj):
            raise TypeError("obj parameter must be a sequence or a map")

        obj = self._encode(obj)
        for k, v in obj.iteritems():
            self._setOneEnv(k, v)
        return obj

    def setEnv(self, key, value):
        """Sets the environment key to the new value and stores it
        persistently.

        :param key: the key for the environment
        :param value: the value for the environment

        :return: a tuple with the key and value objects stored"""
        ret = self.setEnvObj((key, value))
        return key, ret[key]

    def unsetEnv(self, key):
        """Unsets the environment for the given key.

        :param key: the key for the environment to be unset
        :return: the sequence of keys which have been removed"""
        if isinstance(key, (str, unicode)):
            key = (key,)
        self._unsetEnv(key)
        return key



##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from sardana import State
from sardana.pool.controller import IORegisterController


class DummyIORController(IORegisterController):
    """This controller offers as many IORegisters as the user wants (up to 
    1024)."""

    gender = "Simulation"
    model  = "Basic"
    organization = "Sardana team"

    MaxDevice = 1024
    
    predefined_values = "0", "Online", "1" , "Offline", "2", "Standby"
    
    def __init__(self, inst, props, *args, **kwargs):
        IORegisterController.__init__(self, inst, props, *args, **kwargs)
        self.myvalue = 1

    def AddDevice(self, axis):
        self._log.debug('AddDevice %d' % axis)

    def DeleteDevice(self, axis):
        pass

    def StateOne(self, axis):
        return (State.On,"Device in On state")

    def ReadOne(self, axis):
        return self.myvalue

    def WriteOne(self, axis, value):
        self.myvalue = value
        
    def SendToCtrl(self,in_data):
        return ""



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the main macro server module"""

__all__ = ["AttributeLogHandler", "AttributeBufferedLogHandler"]

__docformat__ = 'restructuredtext'

import logging
import weakref
import operator

from taurus.core.util import LIFO


class AttributeLogHandler(logging.Handler):
    
    def __init__(self, dev, attr_name, level=logging.NOTSET, max_buff_size=0):
        logging.Handler.__init__(self, level)
        self._attr_name = attr_name
        self._level = level
        self._max_buff_size = max_buff_size
        self._dev = weakref.ref(dev)
        self._attr = dev.get_device_attr().get_attr_by_name(attr_name)
        self._buff = LIFO(max_buff_size)

    def emit(self, record):
        output = self.getRecordMessage(record)
        self.appendBuffer(output)
        self.sendText(output)
    
    def getRecordMessage(self, record):
        return self.format(record).split('\n')
    
    def sendText(self, output):
        dev = self._dev()
        attr = self._attr
        if attr is None or dev is None:
            return
        dev.set_attribute(attr, output)
    
    def read(self, attr):
        """Read from the buffer and assign to the attribute value"""
        attr.set_value(self._buff.getCopy())
    
    def clearBuffer(self):
        self._buff.clear()
        
    def appendBuffer(self, d):
        if operator.isSequenceType(d):
            if isinstance(d, (str, unicode)):
                self._buff.append(d)
            else:
                self._buff.extend(d)
        else:
            self._buff.append(str(d))
    
    def sync(self):
        pass
    
    def finish(self):
        pass
    
AttributeBufferedLogHandler = AttributeLogHandler


#!/usr/bin/env python

""" The sardana execution tool.
    Syntax:
        python sardana.py [-v] [-l] [--simulation=<mode>] [--cleanup=<True|False>] <file.xml>
        
    This tool creates and executes a sardana system from the given file.
    
    file.xml can be either:
        - a valid XML file coming from an EXCEL spreadsheet.
        - a valid XML file from the sardana schema.
        
    -v  prints an output of the generated XML file and exits
    -l  if given, this parameter activates logging
    
    --simulation=<mode>  supported modes are 'Basic', 'Best' or 'Off'. Default mode is Best.
                         Basic mode will create Dummy* controllers on the Pool
                         Best mode will create Simu* controllers and Simu like device servers
                         Off mode will create the original controllers on the Pool
    WARNING: Any other string will deactivate simulation mode and use REAL mode
    
    --cleanup=<True|False> either or not to perform a cleanup at the end. Default
                           is True.
                           If cleanup is True the database will be cleaned at 
                           shutdown time.
                           Valid True values are: 'yes', 'Yes', 'true', 'True', 'y', 'Y', '1'
                           Any other value will be interpreted as False
"""

import PyTango
import sys
import os
import time, datetime
import exceptions
import imp
import traceback
import logging
import types, operator
import json

from taurus.core.util import CodecFactory

LOG = logging.getLogger()

try:
    from lxml import etree
    LOG.info("Using lxml XML library")
except ImportError:
    try:
        # Python 2.5
        import xml.etree.cElementTree as etree
        LOG.info("Using python native cElemenTree XML library")
    except ImportError:
        try:
            # Python 2.5
            import xml.etree.ElementTree as etree
            LOG.info("Using python native ElemenTree XML library")
        except ImportError:
            try:
                # normal cElementTree install
                import cElementTree as etree
                LOG.info("Using python normal cElemenTree XML library")
            except ImportError:
                try:
                    # normal ElementTree install
                    import elementtree.ElementTree as etree
                    LOG.info("Using python normal ElemenTree XML library")
                except ImportError:
                    LOG.critical("Could not find any suitable XML library")
                    sys.exit(1)
    
try:
    import pexpect
except:
    try:
        import pexpect23 as pexpect
        print "[WARNING]: pexpect module not found. Using local pexpect 2.3"
    except Exception,e:
        print e
        print "The Sardana requires pexpect python module which was not found."
        print "This module can be found at http://www.noah.org/wiki/Pexpect" 
        sys.exit(2)


SAR_NS = 'http://sardana.cells.es/client/framework/config'

class Process:

    ReadyMsg        = "Ready to accept request"
    AlreadyRunning  = "This server is already running, exiting!"
    MaxStartupTime  = 30
    MaxShutdownTime = 30

    def __init__(self, executable, args, name = "Process", instance = None, logfile = None, env = None):
        self._start_time = -1
        self._stop_time = -1
        self._process = None
        self._exec = executable
        self._args = args
        self._name = name
        self._instance = instance
        self._env = env
        self._logfile = logfile

    def o(self,m):
        sys.stdout.write(m)
        sys.stdout.flush()
    
    def on(self,m=''):
        self.o(m)
        sys.stdout.write('\n')
        sys.stdout.flush()

    def getInstance(self):
        return self._instance

    def start(self):
        raise RuntimeException("Not allowed to 'start' abstract Process")
        
    def stop(self):
        raise RuntimeException("Not allowed to 'stop' abstract Process")

    def getMaxStartupTime(self):
        return Process.MaxStartupTime

    def getMaxShutdownTime(self):
        return Process.MaxShutdownTime
    
class PEProcess(Process):

    def __init__(self, executable, args, name = "PEProcess", instance = None, logfile = None, env = None):
        Process.__init__(self, executable, args, name, instance, logfile, env)
    
    def start(self):
        self._start_time = datetime.datetime.now()
        self.o("Starting '%s %s' (%s, %s)... " % (self._name, self._instance, self._exec, self._args))
        try:
            self._process = pexpect.spawn(self._exec, args=self._args, logfile=self._logfile, env=self._env)
            #self.o("%s, %s, %s" % (self._exec, self._args, self._logfile))
            #self._process.logfile = sys.stderr
            
            idx = self._process.expect([Process.ReadyMsg, Process.AlreadyRunning,
                                        pexpect.EOF, pexpect.TIMEOUT],
                                        timeout=self.getMaxStartupTime())
        except Exception,e:
            self.on("[FAILED]")
            raise e
            
        if idx == 0:
            started = datetime.datetime.now()
            dt = started - self._start_time
            self.on("(took %s) [DONE]" % str(dt))
            return

        self.on("[FAILED]")

        if idx == 1:
            raise Exception("%s %s already running" % (self._name, self._instance))
        elif idx == 2:
            raise Exception("%s %s terminated unexpectedly" % (self._name, self._instance))
        elif idx == 3:
            raise Exception("%s %s startup time exceeded" % (self._name, self._instance))
    
    def terminate(self):
        if self._process is None:
            raise Exception("No process")
        return self._process.terminate()
    
    def kill(self):
        if not self._process is None:
            self._process.close()
            res = self._process.terminate(force=True)
            self._process = None
        self.on("[DONE]")
        return
    
    def stop(self, max_shutdown_time=None):
        """Stops"""
        self._stop_time = datetime.datetime.now()
        self.o("Stopping '%s %s'... " % (self._name, self._instance))
        
        try:
            
            max_shutdown_time = max_shutdown_time or self.getMaxShutdownTime()
            
            if self._process is None:
                self.on(" (no process) [DONE]")
                return

            if not self._process.isalive():
                self.on(" (already dead) [DONE]")
                return
            
            try:
                res = self.terminate()
                idx = self._process.expect(['Exiting','Exited', pexpect.EOF, pexpect.TIMEOUT],
                                           timeout = max_shutdown_time)
            except Exception,e:
                self.on("[FAILED]")
                raise e
                
            if idx == 0:
                self.o(".")
            elif idx == 1:
                self.on(" (terminated before expected: Exited) [DONE]")
                return
            elif idx == 2:
                self.on(" (terminated before expected: EOF) [DONE]")
                return
            elif idx == 3:
                self.o(" (shutdown time exceeded). Forcing... ")
                self.kill()
                return
                
            try:
                idx = self._process.expect(['Exited', pexpect.EOF, pexpect.TIMEOUT],
                                           timeout = 5)
            except Exception,e:
                self.on("[FAILED]")
                raise e
            
            if idx == 0:
                self.o(".")
            elif idx == 1:
                self.on(" (terminated before expected: EOF) [DONE]")
            elif idx == 2:
                self.on(" (shutdown time exceeded) [DONE]")
                self.kill()
                return

            try:
                idx = self._process.expect([pexpect.EOF, pexpect.TIMEOUT],
                                           timeout = 5)
            except Exception,e:
                self.on("[FAILED]")
                raise e
        
            if idx == 0:
                runtime = self._stop_time - self._start_time
                stopped = datetime.datetime.now() - self._stop_time
                self.on(". (ran for %s ;took %s) [DONE]" % (str(runtime), str(stopped)))
            elif idx == 1:
                self.o(" (shutdown time exceeded). Forcing... ")
                self.kill()

        except KeyboardInterrupt, ki:
            self.o("(Ctrl-C during stop). Forcing... ")
            self.kill()

    def run(self, timeout = 0):
        if not self._process:
            return
        try:
            res = self._process.read_nonblocking(size = 10000, timeout = timeout)
            
            #self.on(" \t[READ %s.%s] {%s}" % (self._name, self._instance, res))
            
        except pexpect.TIMEOUT:
            # no data was available. Don't worry: it just means that in the mean
            # time no data was sent to the output/err by the daemon
            pass
        except pexpect.EOF:
            # The process terminated. Maybe to something about it in the future
            self.on("%s %s terminated before expected: (EOF)" % (self._name, self._instance))
            self._process = None


class PEDeviceServerProcess(PEProcess):

    def __init__(self, executable, args, name = "PEDeviceServerProcess", instance = None, db = None, logfile = None):
        db = db or PyTango.Database()
        env = os.environ.copy()
        env["TANGO_HOST"] = "%s:%s" % (db.get_db_host(), db.get_db_port())
        PEProcess.__init__(self, executable, args, name, instance, logfile, env)

    def getServerName(self):
        return os.path.splitext(os.path.basename(self._exec))[0]

    def terminate(self):
        tango_host_port = self._env['TANGO_HOST']
        dserver_device_name = "%s/dserver/%s/%s" % (tango_host_port, self.getServerName(), self._instance)
        dserver_device = PyTango.DeviceProxy(dserver_device_name)
        dserver_device.command_inout("Kill")
        PEProcess.terminate(self)

class PEPythonDeviceServerProcess(PEDeviceServerProcess):

    def getServerName(self):
        return os.path.splitext(os.path.basename(self._args[0]))[0]


class PESimuMotorProcess(PEPythonDeviceServerProcess):

    def __init__(self, instname, db = None, logfile = None):
        name = "Motor Simulator"
        # Make sure the python device server code is reachable
        try:
            f, fname, desc = imp.find_module('SimuMotorCtrl')
            if f: f.close()
            f, path, desc = imp.find_module('SimuMotor')
            if f: f.close()
        except exceptions.ImportError, e:
            msg = "Could not find %s executable.\n" \
                  "Make sure PYTHONPATH points to the directory(ies) where " \
                  "SimuMotorCtrl.py and SimuMotor.py files are installed" % name
            raise Exception(msg)
        
        fname = os.path.join(fname,'SimuMotorCtrl.py')
        
        PEDeviceServerProcess.__init__(self, "/usr/bin/python", [fname, instname], name, instname, db, logfile)
    
  
class PESimuCounterTimerProcess(PEPythonDeviceServerProcess):

    def __init__(self, instname, db = None, logfile = None):
        name = "Counter/Timer Simulator"
        # Make sure the python device server code is reachable
        try:
            f, fname, desc = imp.find_module('SimuCoTiCtrl')
            if f: f.close()
        except exceptions.ImportError, e:
            msg = "Could not find %s executable.\n" \
                  "Make sure PYTHONPATH points to the directory(ies) where " \
                  "SimuCoTiCtrl.py file is installed" % name
            raise Exception(msg)
        
        fname = os.path.join(fname,'SimuCoTiCtrl.py')
        
        PEDeviceServerProcess.__init__(self, "/usr/bin/python", [fname, instname], name, instname, db, logfile)


class PEPySignalSimulatorProcess(PEPythonDeviceServerProcess):

    def __init__(self, instname, db = None, logfile = None):
        name = "PySignal Simulator"
        # Make sure the python device server code is reachable
        try:
            f, fname, desc = imp.find_module('PySignalSimulator')
            if f: f.close()
        except exceptions.ImportError, e:
            msg = "Could not find %s executable.\n" \
                  "Make sure PYTHONPATH points to the directory where " \
                  "PySignalSimulator.py is installed" % name
            raise Exception(msg)
        
        fname = os.path.join(fname,'PySignalSimulator.py')
        
        PEDeviceServerProcess.__init__(self, "/usr/bin/python", [fname, instname], name, instname, db, logfile)


class PEDevicePoolProcess(PEDeviceServerProcess):

    def __init__(self, instname, db = None, logfile = None):
        name = "Device Pool"
        ex = None
        for path in os.getenv("PATH").split(":"):
            path = os.path.join(path,"Pool")
            if os.path.exists(path):
                ex = path
                break
        
        if ex is None:
            raise Exception("Could not find %s executable" % name)
        
        args = [instname, "--log-level=info"]
        PEDeviceServerProcess.__init__(self, ex, args, name, instname, db, logfile)

        
class PEMacroServerProcess(PEDeviceServerProcess):

    def __init__(self, instname, db = None, logfile = None):
        name = "Macro Server"

        ex = None
        for path in os.getenv("PATH").split(":"):
            path = os.path.join(path,"MacroServer")
            if os.path.exists(path):
                ex = path
                break
        
        if ex is None:
            raise Exception("Could not find %s executable" % name)

        args = [instname, "--log-level=info"]
        PEDeviceServerProcess.__init__(self, ex, args, name, instname, db, logfile)


SimuMotorProcess = PESimuMotorProcess
SimuCounterTimerProcess = PESimuCounterTimerProcess
PySignalSimulatorProcess = PEPySignalSimulatorProcess
DevicePoolProcess = PEDevicePoolProcess
MacroServerProcess = PEMacroServerProcess

#____________________________________________________________________TangoServer
class TangoServer:

    klassName = "DeviceServer"

    def __init__(self, servNode, bl, createProc = False, log = False):
        if servNode is None:
            raise Exception("No XML data")
            
        self._bl = bl
        self._node = servNode
                
        self._tghost = servNode.get("tangoHost") or None
        self._db = bl.getTangoDB(self._tghost)
        
        self._complete_name = servNode.get("serverName")
        self.on("Preparing %s" % self._complete_name)
        self._klass_name, self._inst_name = self._complete_name.split("/")


        ##########################################################################
        # To prevent some disaster, just make sure the user knows that the
        # hostname where the server will be deleted and created.
        server_host = self._db.get_db_host()
        server_port = self._db.get_db_port()
        server_tango_host = server_host+':'+server_port
        pytango_tango_host = PyTango.ApiUtil.get_env_var('TANGO_HOST')
        pytango_host, pytango_port = pytango_tango_host.split(':')
        # Using socket to get ip addresses because one of the two hostnames could have domain
        import socket
        server_host_ip = socket.gethostbyname_ex(server_host)[2][0]
        pytango_host_ip = socket.gethostbyname_ex(pytango_host)[2][0]
        if (server_host_ip != pytango_host_ip) or (server_port != pytango_port):
            print '\t!!! WARNING !!! %s TANGO_HOST is not the PyTango default. You may erase the WRONG sardana definition.' % self._complete_name
            print '\tServer: %s  PyTango: %s' % (server_tango_host , pytango_tango_host)
            ans = raw_input('\tDo you _really_ want to continue? [y|N] ')
            if ans.lower() not in ['y','yes']:
                raise Exception('User cancelled the creation of %s server' % self._complete_name)
        ##########################################################################

        ##########################################################################
        # Before erasing the content in the database, we will also  create a backup
        # of all the tango devices's properties and memorized attributes "a-la jive".
        # There's an script called jive-save-config that given the parameters
        # <server_name> and <file_name> it saves the config into the specified file the
        # same way you can right-click an instance within jive and select the
        # option 'Save server data'.
        try:
            config_file_name = self._klass_name+'-'+self._inst_name+'-'+time.strftime('%Y%m%d_%H%M%S')+'.jive'
            cmd = 'TANGO_HOST=%s jive-save-config %s %s &>/dev/null' % (server_tango_host, self._complete_name, config_file_name)
            os.system(cmd)
            print 'There is a backup of the deleted server config in: %s' % config_file_name
        except:
            pass
        ##########################################################################
        

        if self.klassName != self._klass_name:
            raise Exception("Node name differs from expeced '%s' != '%s'" % (self._klass_name, self.klassName))
        
        if log:
            self._logfile = open("Log_%s_%s" % (self.klassName, self._inst_name), "w")
        else:
            self._logfile = None
            
        if createProc == True:
            self._proc = self._createProcess()
        else:
            self._proc = None

    def o(self,m):
        sys.stdout.write(m)
        sys.stdout.flush()
    
    def on(self,m=''):
        self.o(m)
        sys.stdout.write('\n')
        sys.stdout.flush()

    def setUp(self):
        """Default setUp."""
        self.prepare()
        self.start()
        
    def tearDown(self):
        """Default tearDown."""
        self.stop()
        self.cleanUp()

    def prepare(self):
        """Prepares everything to be run"""
        if not self._proc:
            self._proc = self._createProcess()

        # Cleanup the database
        self.deleteServerFromDB()
        
        # Create a new instance in database
        self.createServerInDB()

    def deleteServerFromDB(self):
        db = self._db
        server_name = self._klass_name.lower()
        server_instance = self._inst_name.lower()
        server = self._complete_name.lower()
        servers = [ s.lower() for s in db.get_server_list(server_name + '/*') ]
        if server in servers:
            devices = db.get_device_name(server, '*')
            
            for d in devices:
                if not d.startswith('dserver'):
                    props = db.get_device_property_list(d, '*')
                    db.delete_device_property(d, props)
                    db.delete_device(d)
            
            db.delete_server(server)
        
    def createServerInDB(self):
        for devNodeName in self.getDeviceNodeNames():
            
            nodes = self._node.findall(devNodeName)
            for node in nodes:
                
                dev_name = node.get("deviceName")
                if not dev_name:
                    raise Exception("%s does not have valid deviceName" % self._node)
                            
                # Create the device in the database 
                info = PyTango.DbDevInfo()
                info.name   = dev_name
                info._class = devNodeName
                info.server = self._complete_name
                self._db.add_device(info)
                
                alias = node.get("name")
                if alias:
                    self._db.put_device_alias(dev_name, alias)
                 
                props = node.findall("Property")
                if props:
                    self._putDeviceProps(dev_name, props)

    def getDeviceNodeNames(self):
        return [self._klass_name]
    
    def start(self):
        """Default start"""
        self._proc.start()
    
    def stop(self):
        """Default stop"""
        self._proc.stop()
        
    def cleanUp(self):
        """Clean up"""
        self.deleteServerFromDB()
        
    def run(self, step=True, timeout=0):
        """Run"""
        while 1:
            self._proc.run(timeout=timeout)
            if step: return
    
    def _createProcess(self):
        raise Exception("Must overwrite in subclass")

    def _putDeviceProps(self, dev_name, prop_node_list):
        props = {} 
        for p_node in prop_node_list:
            prop_name = p_node.get("name")
            data = []
            
            i_nodes = p_node.findall("Item")
            if len(i_nodes) > 0:
                for i_node in i_nodes:
                    data.append(i_node.text)
            else:
                data.append(p_node.text)
            props[prop_name] = data
        
        if props:
            self._db.put_device_property(dev_name, props)

    def getInstanceName(self):
        return self._inst_name

    def __str__(self):
        return self._complete_name


class SimuMotorServer(TangoServer):
    
    klassName = "SimuMotorCtrl"
    
    def _createProcess(self):
        return SimuMotorProcess(self._inst_name, self._db, logfile=self._logfile)

    def getDeviceNodeNames(self):
        return TangoServer.getDeviceNodeNames(self) + ["SimuMotor"]

        
class SimuCounterTimerServer(TangoServer):
    
    klassName = "SimuCoTiCtrl"
    
    def _createProcess(self):
        return SimuCounterTimerProcess(self._inst_name, self._db, logfile=self._logfile)

    def getDeviceNodeNames(self):
        return TangoServer.getDeviceNodeNames(self) + ["SimuCounter"]


class PySignalSimulatorServer(TangoServer):
    klassName = "PySignalSimulator"
    
    def _createProcess(self):
        return PySignalSimulatorProcess(self._inst_name, self._db, logfile=self._logfile)


class DevicePoolServer(TangoServer):
    
    klassName = "Pool"
    
    def _createProcess(self):
        self.dev_count = 0
        return DevicePoolProcess(self._inst_name, self._db, logfile=self._logfile)

    def _putDeviceProps(self, dev_name, prop_node_list):
        TangoServer._putDeviceProps(self, dev_name, prop_node_list)
        
    def start(self):
        """Default start"""
        self._proc.start()
        self.loadPool()

    def stop(self):
        """Default stop"""
        max_shutdown_time = self.dev_count * 0.2 + 10
        self._proc.stop(max_shutdown_time)

    def _item_node_to_value(self, attr_info, node):
        v = node.text or [i.text or [ j.text for j in i.findall("Item") ] for i in node.findall("Item")]
        return PyTango.seqStr_2_obj(v, attr_info.data_type, attr_info.data_format)

    def handle_attributes(self, dev_name, node):
        attrs = node.findall('Attribute')
        if not len(attrs):
            return
        # Take into account the possibility to have the device in another Tango Host...
        dev_name = self._tghost+'/'+dev_name
        dev = PyTango.DeviceProxy(dev_name)
        tango_attrs = dev.attribute_list_query_ex()
        tango_attrs_map = {}
        for attr in tango_attrs:
            tango_attrs_map[attr.name.lower()] = attr
        
        for attr in attrs:
            name = attr.get("name").lower()
            tango_attr = tango_attrs_map.get(name)
            if tango_attr:
                attr_info = dev.get_attribute_config_ex(name)[0]
                v_node = attr.find("Value")
                if not v_node is None:
                    v = self._item_node_to_value(tango_attr, v_node)
                    try:
                        dev.write_attribute(name, v)
                    except Exception, ex:
                        print 'SOME PROBLEMS SETTING ATTRIBUTE VALUE FOR DEVICE',dev_name,'ATTRIBUTE',tango_attr.name,'VALUE',str(v)
                        print 'EXCEPTION:',ex
                
                c_node = attr.find("Configuration")
                if not c_node is None:
                    disp_node = c_node.find("Display")
                    if not disp_node is None:
                        attr_info.label = disp_node.get("label") or attr_info.label
                        attr_info.format = disp_node.get("format") or attr_info.format
                    units_node = c_node.find("Units")
                    if not units_node is None:
                        attr_info.unit = units_node.get("unit") or attr_info.unit
                        attr_info.display_unit = units_node.get("display_unit") or attr_info.display_unit
                        attr_info.standard_unit = units_node.get("standard_unit") or attr_info.standard_unit
                    range_node = c_node.find("Range")
                    if not range_node is None:
                        attr_info.min_value = range_node.get("min") or attr_info.min_value
                        attr_info.max_value = range_node.get("max") or attr_info.max_value
                    alarms_node = c_node.find("Alarms")
                    if not alarms_node is None:
                        attr_info.alarms.min_warning = alarms_node.get("min_warning") or attr_info.alarms.min_warning
                        attr_info.alarms.max_warning = alarms_node.get("max_warning") or attr_info.alarms.max_warning
                        attr_info.alarms.min_alarm = alarms_node.get("min_alarm") or attr_info.alarms.min_alarm
                        attr_info.alarms.max_alarm = alarms_node.get("max_alarm") or attr_info.alarms.max_alarm
                
                    e_node = attr.find("Events")
                    if not e_node is None:
                        ch_node = e_node.find("ChangeEvent")
                        if not ch_node is None:
                            attr_info.events.ch_event.abs_change = ch_node.get("absolute") or attr_info.events.ch_event.abs_change
                            rel = ch_node.get("relative")
                            if rel:
                                rel = rel.rstrip('%')
                                attr_info.events.ch_event.rel_change = rel
                    
                    p_node = attr.find("Polling")
                    if not p_node is None:
                        polled = p_node.get("polled") or 'False'
                        polled = not (polled.lower() in ('false','no', 'n', '0'))
                        if polled:
                            try:
                                period = int(p_node.get("period") or 0)
                                dev.poll_attribute(name, period)
                            except:
                                print dev_name,tango_attr.name
    
                    try:
                        dev.set_attribute_config(attr_info)
                    except Exception,e:
                        print 'COULD NOT SET THE FOLLOWING CONFIG FOR DEVICE',dev_name,'ATTR', tango_attr.name
                        print 'ATTRIBUTE INFO:',attr_info
                        print 'EXCEPTION:',e

        intrument_node = node.find("InstrumentRef")        
        if not intrument_node is None:
            try:
                value = intrument_node.text.strip()
                dev.write_attribute('Instrument', value)
            except Exception, ex:
                print 'SOME PROBLEMS SETTING INSTRUMENT VALUE FOR DEVICE',dev_name,'VALUE',value
                print 'EXCEPTION:',ex

                
    def loadPool(self):
        start_load_time = datetime.datetime.now()
        self.on("  Loading 'Device Pool %s'..." % self._inst_name)        

        try:
            tgHost = self._node.get("tangoHost") or \
                     ("%s:%s" % (self._db.get_db_host(), self._db.get_db_port()))
            
            pool = self._node.find(self.klassName)
            
            pool_dev_name = "%s/%s" % (tgHost, pool.get("deviceName"))
            
            pool_dp = PyTango.DeviceProxy(pool_dev_name)
            
            factory = CodecFactory()
            elements = factory.decode(pool_dp.elements, ensure_ascii='True')['new']
            
            ctrl_classes_info = {}
            for elem in elements:
                if elem['type'] != 'ControllerClass':
                    continue
                ctrl_classes_info[elem['name']] = elem

            count = 0           
            # Setup instruments
            instruments = pool.xpath("Instrument")
            self.o("    Creating instruments ")
            for instrument in instruments:
                try:
                    name = instrument.get("name")
                    if name.startswith('#'):
                        self.o("x")
                        continue
                    if name.startswith('REAL'):
                        self.o('R')
                    else:
                        self.o(".")
                    kclass = instrument.get("class")
                    pars = name, kclass
                    pool_dp.command_inout("CreateInstrument", pars)
                    count += 1
                except:
                    self.on("[FAILED]")
                    raise
            self.on(" (%d) [DONE]" % count)

            ctrls = pool.xpath("Controller[@type != 'PseudoMotor' and @type != 'PseudoCounter']")
            self.o("    Creating controllers ")
            count = 0
            self.dev_count = 0            
            for ctrl in ctrls:
                try:
                    name = ctrl.get("name")
                    # skip controllers which name starts with '#'
                    if name.startswith('#'):
                        self.o("x")
                        continue
                    if name.startswith('REAL'):
                        self.o('R')
                    else:
                        self.o(".")
                    type = ctrl.get("type")
                    if type == 'CounterTimer':
                        type = 'CTExpChannel'
                    lib = ctrl.get("lib")
                    kclass = ctrl.get("class")
                    pars = [type, lib, kclass, name]
                    props = ctrl.findall("Property")
                    for p in props:
                        pars.append(p.get("name"))
                        pars.append(p.text or '\n'.join([i.text for i in p.findall("Item")]))
                    pars = map(str.strip, pars)
                    pool_dp.command_inout("CreateController", pars)
                    self.run(step = True) # to flush any output generated by the pool
                    count += 1
                except:
                    self.on("[FAILED]")
                    raise
            self.on(" (%d) [DONE]" % count)
            
            ctrls = pool.xpath("Controller[@type = 'Motor']")
            self.o("    Creating motors ")
            count = 0
            for ctrl in ctrls:
                name = ctrl.get("name")
                elems = ctrl.findall("Motor")
                # skip motor creation for controllers which name starts with '#'
                if name.startswith('#'):
                    elems = []
                for e in elems:
                    try:
                        axis = e.get("axis")
                        aliasName = e.get("name")
                        deviceName = e.get("deviceName") or ""
                        # skip motor creation for alias starting with '#'
                        if aliasName.startswith('#'):
                            self.o('x')
                            continue
                        self.o(".")
                        #pars = [ [axis], [ aliasName, name ] ]
                        pars = "Motor", name, axis, aliasName
                        if deviceName.count('/') == 2:
                            pars.append(deviceName)
                        pars = map(str.strip, pars)
                        pool_dp.command_inout("CreateElement", pars)

                        self.handle_attributes(aliasName, e)
                        
                        self.run(step = True) # to flush any output generated by the pool
                        count += 1
                        self.dev_count += 1
                    except:
                        self.on("[FAILED]")
                        raise
            self.on(" (%d) [DONE]" % count)
            
            ctrls = pool.xpath("Controller[@type = 'CounterTimer']")
            self.o("    Creating Counter/Timers ")
            count = 0
            for ctrl in ctrls:
                name = ctrl.get("name")
                elems = ctrl.findall("CounterTimer")
                # skip counter/timers creation for controllers which name starts with '#'
                if name.startswith('#'):
                    elems = []
                for e in elems:
                    try:
                        axis = e.get("axis")
                        aliasName = e.get("name")
                        deviceName = e.get("deviceName") or ""
                        # skip counter/timers creation for alias starting with '#'
                        if aliasName.startswith('#'):
                            self.o('x')
                            continue
                        self.o(".")
                        pars = "CTExpChannel", name, axis, aliasName
                        if deviceName.count('/') == 2:
                            pars.append(deviceName)
                        pars = map(str.strip, pars)
                        pool_dp.command_inout("CreateElement", pars)

                        self.handle_attributes(aliasName, e)
                        
                        self.run(step = True) # to flush any output generated by the pool
                        count += 1
                        self.dev_count += 1
                    except:
                        self.on("[FAILED]")
                        raise
            self.on(" (%d) [DONE]" % count)
                    
            ctrls = pool.xpath("Controller[@type = 'ZeroDExpChannel']")
            self.o("    Creating 0Ds ")
            count = 0
            for ctrl in ctrls:
                name = ctrl.get("name")
                elems = ctrl.findall("ZeroDExpChannel")
                # skip 0Ds creation for controllers which name starts with '#'
                if name.startswith('#'):
                    elems = []
                for e in elems:
                    try:
                        axis = e.get("axis")
                        aliasName = e.get("name")
                        deviceName = e.get("deviceName") or ""
                        # skip 0Ds creation for alias starting with '#'
                        if aliasName.startswith('#'):
                            self.o('x')
                            continue
                        self.o(".")
                        pars = "ZeroDExpChannel", name, axis, aliasName
                        if deviceName.count('/') == 2:
                            pars.append(deviceName)
                        pars = map(str.strip, pars)
                        try:
                            pool_dp.command_inout("CreateElement", pars)

                            self.handle_attributes(aliasName, e)

                        except PyTango.DevFailed, df:
                            self.on("Exception creating %s: %s" %(aliasName, str(df)))
                        self.run(step = True) # to flush any output generated by the pool
                        count += 1
                        self.dev_count += 1
                    except:
                        self.on("[FAILED]")
                        raise
            self.on(" (%d) [DONE]" % count)

            ctrls = pool.xpath("Controller[@type = 'OneDExpChannel']")
            self.o("    Creating 1Ds ")
            count = 0
            for ctrl in ctrls:
                name = ctrl.get("name")
                elems = ctrl.findall("OneDExpChannel")
                # skip 1Ds creation for controllers which name starts with '#'
                if name.startswith('#'):
                    elems = []
                for e in elems:
                    try:
                        axis = e.get("axis")
                        aliasName = e.get("name")
                        deviceName = e.get("deviceName") or ""
                        # skip 1Ds creation for alias starting with '#'
                        if aliasName.startswith('#'):
                            self.o('x')
                            continue
                        self.o(".")
                        pars = "OneDExpChannel", name, axis, aliasName
                        if deviceName.count('/') == 2:
                            pars.append(deviceName)
                        pars = map(str.strip, pars)
                        pool_dp.command_inout("CreateElement", pars)

                        self.handle_attributes(aliasName, e)

                        self.run(step = True) # to flush any output generated by the pool
                        count += 1
                        self.dev_count += 1
                    except:
                        self.on("[FAILED]")
                        raise
            self.on(" (%d) [DONE]" % count)

            ctrls = pool.xpath("Controller[@type = 'TwoDExpChannel']")
            self.o("    Creating 2Ds ")
            count = 0
            for ctrl in ctrls:
                name = ctrl.get("name")
                elems = ctrl.findall("TwoDExpChannel")
                # skip 2Ds creation for controllers which name starts with '#'
                if name.startswith('#'):
                    elems = []
                for e in elems:
                    try:
                        axis = e.get("axis")
                        aliasName = e.get("name")
                        deviceName = e.get("deviceName") or ""
                        # skip 2Ds creation for alias starting with '#'
                        if aliasName.startswith('#'):
                            self.o('x')
                            continue
                        self.o(".")
                        pars = "TwoDExpChannel", name, axis, aliasName
                        if deviceName.count('/') == 2:
                            pars.append(deviceName)
                        pars = map(str.strip, pars)
                        pool_dp.command_inout("CreateElement", pars)

                        self.handle_attributes(aliasName, e)

                        self.run(step = True) # to flush any output generated by the pool
                        count += 1
                        self.dev_count += 1
                    except:
                        self.on("[FAILED]")
                        raise
            self.on(" (%d) [DONE]" % count)
            
            ctrls = pool.xpath("Controller[@type = 'IORegister']")
            self.o("    Creating IORegisters ")
            count = 0
            for ctrl in ctrls:
                name = ctrl.get("name")
                elems = ctrl.findall("IORegister")
                # skip IORegisters creation for controllers which name starts with '#'
                if name.startswith('#'):
                    elems = []
                for e in elems:
                    try:
                        axis = e.get("axis")
                        aliasName = e.get("name")
                        deviceName = e.get("deviceName") or ""
                        # skip IORegisters creation for alias starting with '#'
                        if aliasName.startswith('#'):
                            self.o('x')
                            continue
                        self.o(".")
                        pars = "IORegister", name, axis, aliasName
                        if deviceName.count('/') == 2:
                            pars.append(deviceName)
                        pars = map(str.strip, pars)
                        pool_dp.command_inout("CreateElement", pars)

                        self.handle_attributes(aliasName, e)
                        
                        self.run(step = True) # to flush any output generated by the pool
                        count += 1
                        self.dev_count += 1
                    except:
                        self.on("[FAILED]")
                        raise
            self.on(" (%d) [DONE]" % count)
            
            #-------------------------------------------------
            # add controllers that depend on physical elements
            #-------------------------------------------------
            ctrls = pool.xpath("Controller[@type = 'PseudoMotor']")
            
            self.o("    Creating Pseudo Motors ")
            pm_ctrl_count = 0
            pm_count = 0
            for ctrl in ctrls:
                try:
                    name = ctrl.get("name")
                    # skip controllers which name starts with '#'
                    if name.startswith('#'):
                        self.o("x")
                        continue
                    self.o(".")
                    type = ctrl.get("type")
                    lib = ctrl.get("lib")
                    kclass = ctrl.get("class")
                    
                    ctrl_class_info = ctrl_classes_info[kclass]                    
                    motor_roles = ctrl_class_info['motor_roles']
                    pseudo_motor_roles = ctrl_class_info['pseudo_motor_roles']
                    pars = [type, lib, kclass, name]
                    
                    for i, e in enumerate(ctrl.findall("Motor")):
                        pars.append("%s=%s" % (motor_roles[i], e.text)) 
                                             
                    for i, e in enumerate(ctrl.findall("PseudoMotor")):
                        pars.append("%s=%s" % (pseudo_motor_roles[i], e.get('name')))

                    pm_count += len(ctrl.findall("PseudoMotor"))
                    self.dev_count += pm_count
                        
                    props = ctrl.findall("Property")
                    
                    for p in props:
                        pars.append(p.get("name"))
                        pars.append(p.text or '\n'.join([i.text for i in p.findall("Item")]))

                    pars = map(str.strip, pars)

                    pool_dp.command_inout("CreateController", pars)
                    pm_ctrl_count += 1

                    for e in ctrl.findall('PseudoMotor'):
                        self.handle_attributes(e.get("name"), e)
                    
                    self.run(step = True) # to flush any output generated by the pool
                except:
                    self.on("[FAILED]")
                    print ctrl_class_info
                    raise
            self.on(" (%d ctrls; %d pmotors) [DONE]" % (pm_ctrl_count, pm_count))
            
            
            
            ctrls = pool.xpath("Controller[@type = 'PseudoCounter']")
            
            self.o("    Creating Pseudo Counters ")
            pc_ctrl_count = 0
            pc_count = 0
            for ctrl in ctrls:
                try:
                    name = ctrl.get("name")
                    # skip controllers which name starts with '#'
                    if name.startswith('#'):
                        self.o("x")
                        continue
                    self.o(".")
                    type = ctrl.get("type")
                    lib = ctrl.get("lib")
                    kclass = ctrl.get("class")

                    ctrl_class_info = ctrl_classes_info[kclass]                    
                    counter_roles = ctrl_class_info['counter_roles']
                    pseudo_counter_roles = ctrl_class_info['pseudo_counter_roles']

                    pars = [type, lib, kclass, name]

                    for i, e in enumerate(ctrl.findall("Channel")):
                        pars.append("%s=%s" % (counter_roles[i], e.text))
                                             
                    for i, e in enumerate(ctrl.findall("PseudoCounter")):
                        pars.append("%s=%s" % (pseudo_counter_roles[i], e.get('name'))) 

                    pc_count += len(ctrl.findall("PseudoCounter"))
                    self.dev_count += pc_count
                        
                    props = ctrl.findall("Property")
                    
                    for p in props:
                        pars.append(p.get("name"))
                        pars.append(p.text or '\n'.join([i.text for i in p.findall("Item")]))

                    pars = map(str.strip, pars)

                    pool_dp.command_inout("CreateController", pars)
                    pc_ctrl_count += 1

                    for e in ctrl.findall('PseudoCounter'):
                        self.handle_attributes(e.get("name"), e)
                    
                    self.run(step = True) # to flush any output generated by the pool
                except:
                    self.on("[FAILED]")
                    raise
            self.on("(%d ctrls; %d pcounters) [DONE]" % (pc_ctrl_count, pc_count))

            measurement_groups = pool.findall("MeasurementGroup")
            self.o("    Creating Measurement Groups ")
            count = 0
            for mg in measurement_groups:
                try:
                    aliasName = mg.get("name")
                    # skip measurementgroups which name starts with '#'
                    if aliasName.startswith('#'):
                        self.o("x")
                        continue
                    self.o(".")

                    channels = mg.findall("ChannelRef")
                    pars = [ aliasName ]
                    for channel in channels:
                        pars.append(channel.get("name"))
                    pars = map(str.strip, pars)
                    pool_dp.command_inout("CreateMeasurementGroup", pars)
                    
                    self.handle_attributes(aliasName, mg)
                    
                    self.run(step = True) # to flush any output generated by the pool
                    count += 1
                    self.dev_count += 1
                except:
                    self.on("[FAILED]")
                    raise
            self.on(" (%d) [DONE]" % count)
     
            end_load_time = datetime.datetime.now()
            dt = end_load_time - start_load_time
            self.on("  Loading 'Device Pool %s'... (took %s) [DONE]" % (self._inst_name, str(dt)))
        except:
            self.on("  Loading 'Device Pool %s'... [FAILED]" % self._inst_name)
            raise


class MacroServerServer(TangoServer):
    
    klassName = "MacroServer"
    
    def _createProcess(self):
        return MacroServerProcess(self._inst_name, self._db, logfile=self._logfile)

    def getDeviceNodeNames(self):
        return TangoServer.getDeviceNodeNames(self) + ["Door"]

    def start(self):
        try:
            TangoServer.start(self)
        except:
            pass
              
                
class Sardana:  
    """Generic Sardana system"""
    
    SimulationModes = { 
        "Basic" : {           "Motor" : ("DummyMotorController.py", "DummyMotorController"),
                       "CounterTimer" : ("DummyCounterTimerController.py", "DummyCounterTimerController"),
                    "ZeroDExpChannel" : ("DummyZeroDController.py", "DummyZeroDController"),
                     "OneDExpChannel" : ("Dummy1DController.py", "'Dummy1DController"),
                     "TwoDExpChannel" : ("Dummy2DController.py", "'Dummy2DController"),
                         "IORegister" : ("DummyIORController.py", "DummyIORController") },
        "Best" :  {           "Motor" : ("SimuMotCtrl.py", "SimuMotorController"),
                       "CounterTimer" : ("SimuCTCtrl.py", "SimuCoTiController"),
                    "ZeroDExpChannel" : ("Simu0DCtrl.py", "Simu0DController"),
                     "OneDExpChannel" : ("Simu1DCtrl.py", "Simu1DController"),
                     "TwoDExpChannel" : ("Simu2DCtrl.py", "Simu2DController"),
                         "IORegister" : ("SimuIOCtrl.py", "SimuIOController"), },
        "Off" :   {           "Motor" : (None,None),
                       "CounterTimer" : (None,None),
                    "ZeroDExpChannel" : (None,None),
                     "OneDExpChannel" : (None,None),
                     "TwoDExpChannel" : (None,None),
                         "IORegister" : (None,None) },
    }
    
    def __init__(self, source, simulation="Best", cleanup=True, log=False):
    
        if type(source) in types.StringTypes:
            self._filename = source
            self._xmldoc = None
        else:
            self._xmldoc = source
            self._filename = self._xmldoc.docinfo.URL
        self._cleanup = cleanup
        self._dft_tg_db = None
        self._tg_dbs = {}
        self._simulation = simulation
        self._log = log

    def o(self,m):
        sys.stdout.write(m)
        sys.stdout.flush()
    
    def on(self,m=''):
        self.o(m)
        sys.stdout.write('\n')
        sys.stdout.flush()

    def getTangoDB(self, id = None):
        if id is None:
            if self._dft_tg_db is None:
                db = PyTango.Database()
                self._dft_tg_db = db
                dft_host = "%s:%s" % (db.get_db_host(), db.get_db_port())
                self._tg_dbs[dft_host] = db
            return self._dft_tg_db
            
        db = self._tg_dbs.get(id)
        if not db is None:
            return db
        try:
            id = str(id)
            host,port = id.split(":")
            port = int(port)
            db = PyTango.Database(host,port)
            self._tg_dbs[id] = db
            return db
        except:
            return None
                    
    def setUp(self):
        """Default setUp."""
        self.prepare()
        self.start()
        
    def tearDown(self):
        """Default tearDown."""
        self.stop()
        if self._cleanup:
            self.cleanUp()
    
    def _preprocess(self):
        start_pp_time = datetime.datetime.now()
        self.o("Preprocessing input... ")
        
        try:
            if not self.SimulationModes.has_key(self._simulation):
                return

            motSimNb   = 0
            ctSimNb    = 0
            zeroDSimNb = 0
            oneDSimNb  = 0
            twoDSimNb  = 0
            ioSimNb    = 0
            pySigSimNb = 0
                        
            sarName = self._sarNode.get("name")
            simType = self.SimulationModes.get(self._simulation)
            poolservers = self._sarNode.findall("PoolServer")
            
            SimuMotorLib, SimuMotorClass = simType["Motor"]
            SimuCoTiLib, SimuCoTiClass = simType["CounterTimer"]
            Simu0DLib, Simu0DClass = simType["ZeroDExpChannel"]
            Simu1DLib, Simu1DClass = simType["OneDExpChannel"]
            Simu2DLib, Simu2DClass = simType["TwoDExpChannel"]
            SimuIOLib, SimuIOClass = simType["IORegister"]
            
            simuMotorLibs = [ mode["Motor"][0] for mode in self.SimulationModes.values()]
            simuCoTiLibs = [ mode["CounterTimer"][0] for mode in self.SimulationModes.values()]
            simu0DLibs = [ mode["ZeroDExpChannel"][0] for mode in self.SimulationModes.values()]
            simu1DLibs = [ mode["OneDExpChannel"][0] for mode in self.SimulationModes.values()]
            simu2DLibs = [ mode["TwoDExpChannel"][0] for mode in self.SimulationModes.values()]
            simuIOLibs = [ mode["IORegister"][0] for mode in self.SimulationModes.values()]
            
            for poolserver in poolservers:
                simuMotorList = []
                pool = poolserver.find("Pool")

                # Process Motor controllers
                motctrls = pool.xpath("Controller[@type = 'Motor']")
                needsServerNumber = len(motctrls) > 1
                for ctrl in motctrls:
                    lib = ctrl.get("lib")
                    klass = ctrl.get("class")
                    
                    if self._simulation == "Off" or lib in simuMotorLibs or ctrl.get('name').startswith('REAL'):
                        continue
            
                    motSimNb += 1
                    
                    if self._simulation == "Best":
                        # add a SimuMotorServer node to the Sardana node
                        simuServer = etree.SubElement(self._sarNode, "SimuMotorServer")
                        tgHost = poolserver.get("tangoHost")
                        if tgHost: 
                            simuServer.set("tangoHost", tgHost)
                        serverName = "SimuMotorCtrl/%s" % sarName
                        if needsServerNumber:
                            serverName += "%03d" % motSimNb
                        simuServer.set("serverName", serverName)
                        
                        simuMotCtrl = etree.SubElement(simuServer, "SimuMotorCtrl")
                        simuMotCtrlName = "%s/simumotctrl/%03d" % (sarName, motSimNb)
                        simuMotCtrl.set("deviceName", simuMotCtrlName)
                        maxError = etree.SubElement(simuMotCtrl, "Property")
                        maxError.set("name","MaxError")
                        maxErrorV = etree.SubElement(maxError, "Item")
                        maxErrorV.text = "0.0"
                        
                        for m in ctrl.findall("Motor"):
                            m_nb = int(m.get('axis'))
                            simuMotor = etree.SubElement(simuServer, "SimuMotor")
                            devName = "%s/simumotctrl%03d/%03d" % (sarName, motSimNb, m_nb)
                            simuMotor.set("deviceName", devName)
                            simuMotor.set("class", "SimuMotor")
                            simuMotorList.append(simuMotor)
                    
                    # change the pool XML nodes to refer to simulator lib
                    # instead of real lib
                    map(ctrl.remove, ctrl.findall("Property"))
                    ctrl.set("lib", SimuMotorLib)
                    ctrl.set("class", SimuMotorClass)
                    
                    if self._simulation == "Best":
                        devName = etree.SubElement(ctrl, "Property")
                        devName.set("name", "DevName")
                        devNameV = etree.SubElement(devName, "Item")
                        devNameV.text = simuMotCtrlName
                
                # Process CounterTimer controllers
                ctctrls = pool.xpath("Controller[@type = 'CounterTimer' and " \
                                                "@lib != '%s' and @class != '%s']" \
                                                 % ("UxTimerCtrl.la", "UnixTimer"))
                needsServerNumber = len(ctctrls) > 1
                for ctrl in ctctrls:
                    lib = ctrl.get("lib")
                    klass = ctrl.get("class")
                    
                    if self._simulation == "Off" or lib in simuCoTiLibs or ctrl.get('name').startswith('REAL'):
                        continue
                    
                    ctSimNb += 1
                    
                    if self._simulation == "Best":
                        # add a SimuCoTiServer node to the Sardana node
                        simuServer = etree.SubElement(self._sarNode, "SimuCoTiServer")
                        tghost = poolserver.get("tangoHost")
                        if tghost: 
                            simuServer.set("tangoHost", tgHost)
                        serverName = "SimuCoTiCtrl/%s" % sarName
                        if needsServerNumber:
                            serverName += "%03d" % ctSimNb
                        simuServer.set("serverName", serverName)
                        
                        simuCoTiCtrl = etree.SubElement(simuServer,"SimuCoTiCtrl")
                        simuCoTiCtrlName = "%s/simuctctrl/%03d" % (sarName, ctSimNb)
                        simuCoTiCtrl.set("deviceName", simuCoTiCtrlName)
                        
                        motRef = 0
                        for ct in ctrl.findall("CounterTimer"):
                            ct_nb = int(ct.get('axis'))
                            simuCounter = etree.SubElement(simuServer, "SimuCounter")
                            devName = "%s/simuctctrl%03d/%03d" % (sarName, ctSimNb, ct_nb)
                            simuCounter.set("deviceName", devName)
                            simuCounter.set("class", "SimuCounter")
                            pAverage = etree.SubElement(simuCounter, "Property")
                            pAverage.set("name", "Average")
                            pAverage.set("type", "DevDouble")
                            pAverage.text = "50.0"
                    
                            pMax = etree.SubElement(simuCounter, "Property")
                            pMax.set("name", "Max")
                            pMax.set("type", "DevDouble")
                            pMax.text = "500.0"

                            pSigma = etree.SubElement(simuCounter, "Property")
                            pSigma.set("name", "Sigma")
                            pSigma.set("type", "DevDouble")
                            pSigma.text = "250.0"
                        
                            if simuMotorList:
                                pMotorName = etree.SubElement(simuCounter, "Property")
                                pMotorName.set("name", "MotorName")
                                pMotorName.set("type", "DevString")
                                pMotorName.text = simuMotorList[motRef].get("deviceName")
                                if motRef == len(simuMotorList) - 1:
                                    motRef = 0
                                else:
                                    motRef += 1
                        
                    # change the pool XML nodes to refer to simulator lib
                    # instead of real lib
                    map(ctrl.remove, ctrl.findall("Property"))
                    ctrl.set("lib", SimuCoTiLib)
                    ctrl.set("class", SimuCoTiClass)
                    if self._simulation == "Best":
                        devName = etree.SubElement(ctrl, "Property")
                        devName.set("name", "DevName")
                        devNameV = etree.SubElement(devName, "Item")
                        devNameV.text = simuCoTiCtrlName
                
                pySigSimNode = None
                
                # Process 0D controllers
                zerodctrls = pool.xpath("Controller[@type = 'ZeroDExpChannel']")
                for ctrl in zerodctrls:
                    lib = ctrl.get("lib")
                    klass = ctrl.get("class")
                    
                    if self._simulation == "Off" or lib in simu0DLibs or ctrl.get('name').startswith('REAL'):
                        continue
                    
                    zeroDSimNb += 1
                    pySigSimNb += 1
                    
                    if self._simulation == "Best":
                        if pySigSimNode is None:
                            # add a PySignalSimulator node to the Sardana node
                            pySigSimNode = etree.SubElement(self._sarNode, "PySignalSimulatorServer")
                            tghost = poolserver.get("tangoHost")
                            if tghost: 
                                pySigSimNode.set("tangoHost", tgHost)
                            serverName = "PySignalSimulator/%s" % sarName
                            pySigSimNode.set("serverName", serverName)
                        
                        simu0DCtrl = etree.SubElement(pySigSimNode, "PySignalSimulator")
                        simu0DCtrlName = "%s/PySignalSimulator/%03d" % (sarName, pySigSimNb)
                        simu0DCtrl.set("deviceName", simu0DCtrlName)
                        
                        zerods = ctrl.findall("ZeroDExpChannel")
                        
                        if len(zerods) > 0:
                            pSimAttributes = etree.SubElement(simu0DCtrl, "Property")
                            pSimAttributes.set("name", "DynamicAttributes")
                            pSimAttributes.set("type", "DevVarStringArray")
                            simAttrTempl = "zerod%03d=float(100.0+10.0*random())"
                            for i in xrange(len(zerods)):
                                pSimAttributeItem = etree.SubElement(pSimAttributes, "Item")
                                pSimAttributeItem.text = simAttrTempl % (i+1)
                        
                    # change the pool XML nodes to refer to simulator lib
                    # instead of real lib
                    map(ctrl.remove, ctrl.findall("Property"))
                    ctrl.set("lib", Simu0DLib)
                    ctrl.set("class", Simu0DClass)
                    if self._simulation == "Best":
                        attributeNames = etree.SubElement(ctrl, "Property")
                        attributeNames.set("name", "AttributeNames")
                        simAttrTempl = "%s/zerod%%03d" % simu0DCtrlName
                        for i in xrange(len(zerods)):
                            attributeNameItem = etree.SubElement(attributeNames, "Item")
                            attributeNameItem.text = simAttrTempl % (i+1)  

                # Process 1D controllers
                onedctrls = pool.xpath("Controller[@type = 'OneDExpChannel']")
                for ctrl in onedctrls:
                    lib = ctrl.get("lib")
                    klass = ctrl.get("class")
                    
                    if self._simulation == "Off" or lib in simu1DLibs or ctrl.get('name').startswith('REAL'):
                        continue
                    
                    oneDSimNb += 1
                    pySigSimNb += 1
                    
                    if self._simulation == "Best":
                        if pySigSimNode is None:
                            # add a PySignalSimulator node to the Sardana node
                            pySigSimNode = etree.SubElement(self._sarNode, "PySignalSimulatorServer")
                            tghost = poolserver.get("tangoHost")
                            if tghost: 
                                pySigSimNode.set("tangoHost", tgHost)
                            serverName = "PySignalSimulator/%s" % sarName
                            pySigSimNode.set("serverName", serverName)
                        
                        simu1DCtrl = etree.SubElement(pySigSimNode, "PySignalSimulator")
                        simu1DCtrlName = "%s/PySignalSimulator/%03d" % (sarName, pySigSimNb)
                        simu1DCtrl.set("deviceName", simu1DCtrlName)
                        
                        onedds = ctrl.findall("OneDExpChannel")
                        
                        if len(onedds) > 0:
                            pSimAttributes = etree.SubElement(simu1DCtrl, "Property")
                            pSimAttributes.set("name", "DynamicAttributes")
                            pSimAttributes.set("type", "DevVarStringArray")
                            
                            simAttrTempl = "oned%03d=DevVarLongArray([10*sin(0.01*x) for x in xrange(100)])"
                            for i in xrange(len(oneds)):
                                pSimAttributeItem = etree.SubElement(pSimAttributes, "Item")
                                pSimAttributeItem.text = simAttrTempl % (i+1)
                        
                    # change the pool XML nodes to refer to simulator lib
                    # instead of real lib
                    map(ctrl.remove, ctrl.findall("Property"))
                    ctrl.set("lib", Simu1DLib)
                    ctrl.set("class", Simu1DClass)
                    if self._simulation == "Best":
                        attributeNames = etree.SubElement(ctrl, "Property")
                        attributeNames.set("name", "AttributeNames")
                        attributeNamesV = etree.SubElement(attributeNames, "Item")
                        simAttrTempl = "%s/oned%%03d" % simu1DCtrlName
                        for i in xrange(len(oneds)):
                            attributeNameItem = etree.SubElement(attributeNames, "Item")
                            attributeNameItem.text = simAttrTempl % (i+1)  
                        
                # Process IORegister controllers
                ioctrls = pool.xpath("Controller[@type = 'IORegister']")
                for ctrl in ioctrls:
                    lib = ctrl.get("lib")
                    klass = ctrl.get("class")
                    
                    if self._simulation == "Off" or lib in simuIOLibs or ctrl.get('name').startswith('REAL'):
                        continue
                    
                    ioSimNb += 1
                    pySigSimNb += 1
                    
                    if self._simulation == "Best":
                        if pySigSimNode is None:
                            # add a PySignalSimulator node to the Sardana node
                            pySigSimNode = etree.SubElement(self._sarNode, "PySignalSimulatorServer")
                            tghost = poolserver.get("tangoHost")
                            if tghost: 
                                pySigSimNode.set("tangoHost", tgHost)
                            serverName = "PySignalSimulator/%s" % sarName
                            pySigSimNode.set("serverName", serverName)
                        
                        simuIOCtrl = etree.SubElement(pySigSimNode, "PySignalSimulator")
                        simuIOCtrlName = "%s/PySignalSimulator/%03d" % (sarName, pySigSimNb)
                        simuIOCtrl.set("deviceName", simuIOCtrlName)
                        
                        iors = ctrl.findall("IORegister")
                        
                        if len(iors) > 0:
                            pSimAttributes = etree.SubElement(simuIOCtrl, "Property")
                            pSimAttributes.set("name", "DynamicAttributes")
                            pSimAttributes.set("type", "DevVarStringArray")
                            
                            simAttrTempl = "ior%03d=int(READ and VAR('ior%03d') or WRITE and VAR('ior%03d',VALUE))"
                            for i in xrange(len(iors)):
                                pSimAttributeItem = etree.SubElement(pSimAttributes, "Item")
                                pSimAttributeItem.text = simAttrTempl % ((i+1),(i+1),(i+1))
                        
                    # change the pool XML nodes to refer to simulator lib
                    # instead of real lib
                    map(ctrl.remove, ctrl.findall("Property"))
                    ctrl.set("lib", SimuIOLib)
                    ctrl.set("class", SimuIOClass)
                    if self._simulation == "Best":
                        attributeNames = etree.SubElement(ctrl, "Property")
                        attributeNames.set("name", "AttributeNames")
                        simAttrTempl = "%s/ior%%03d" % simuIOCtrlName
                        for i in xrange(len(iors)):
                            attributeNameItem = etree.SubElement(attributeNames, "Item")
                            attributeNameItem.text = simAttrTempl % (i+1)  

            end_pp_time = datetime.datetime.now()
            dt = end_pp_time - start_pp_time
            self.on("(took %s) [DONE]" % str(dt))
        except:
            self.on("[FAILED]")
            raise
        
    def prepare(self):
        """Prepares everything to be run"""
        if not self._xmldoc:
            self._xmldoc = etree.parse(self._filename)
            
        self._sarNode = self._xmldoc.getroot()
        
        sarNodeName = "Sardana"
        
        if self._sarNode.tag != sarNodeName:
            raise Exception("<Sardana> root node not found in %s" % self._filename)

        self._preprocess()

        self.prepareMotorSimulators()
        self.prepareCounterTimerSimulators()
        self.prepareSignalSimulators()
        self.prepareDevicePools()
        self.prepareMSs()
        self.prepareServices()
        
    def prepareServices(self):
        sarName = self._sarNode.get("name")
        macro_servers = self._sarNode.findall("MacroServerServer/MacroServer")
        db = self.getTangoDB()
        
        if len(macro_servers) > 0:
            ms_dev_name = macro_servers[0].get("deviceName")
            db.register_service("Sardana", sarName, ms_dev_name)
            self._service = sarName, ms_dev_name
            return
        
        pools = self._sarNode.findall("Pool") 
        if len(pools) > 0:
            pool_dev_name = pools[0].get("deviceName")
            db.register_service("Sardana", sarName, pool_dev_name)
            self._service = sarName, pool_dev_name
            return
       
    def prepareMotorSimulators(self):
        servNodes = self._sarNode.findall("SimuMotorServer")
        
        self._motorSims = {}
        
        for servNode in servNodes:
            serv = SimuMotorServer(servNode, self, log=self._log)
            serv.prepare()
            self._motorSims[serv.getInstanceName()] = serv
            
    def prepareCounterTimerSimulators(self):
        servNodes = self._sarNode.findall("SimuCoTiServer")
        
        self._coTiSims = {}
        
        for servNode in servNodes:
            serv = SimuCounterTimerServer(servNode, self, log=self._log)
            serv.prepare()
            self._coTiSims[serv.getInstanceName()] = serv

    def prepareSignalSimulators(self):
        servNodes = self._sarNode.findall("PySignalSimulatorServer")
        
        self._signalSims = {}
        
        for servNode in servNodes:
            serv = PySignalSimulatorServer(servNode, self, log=self._log)
            serv.prepare()
            self._signalSims[serv.getInstanceName()] = serv

    def prepareDevicePools(self):
        servNodes = self._sarNode.findall("PoolServer")

        self._devicePools = {}
        
        for servNode in servNodes:
            serv = DevicePoolServer(servNode, self, log=self._log)
            serv.prepare()
            self._devicePools[serv.getInstanceName()] = serv

    def prepareMSs(self):
        servNodes = self._sarNode.findall("MacroServer")
        
        self._macServs = {}
        
        for servNode in servNodes:
            serv = MacroServerServer(servNode, self, log=self._log)
            serv.prepare()
            self._macServs[serv.getInstanceName()] = serv

    def _getServsShutdownOrder(self):
        servs  = self._macServs.values()
        servs += self._devicePools.values()
        servs += self._signalSims.values()
        servs += self._coTiSims.values()
        servs += self._motorSims.values()
        return servs

    def _getServsStartupOrder(self):
        servs  = self._motorSims.values()
        servs += self._coTiSims.values()
        servs += self._signalSims.values()
        servs += self._devicePools.values()
        servs += self._macServs.values()
        return servs
    
    def cleanUp(self):
        for s in self._getServsShutdownOrder():
            s.cleanUp()
        
        if hasattr(self, '_service') and not self._service is None:
            db = self.getTangoDB()
            db.unregister_service("Sardana", self._service[0])
    
    def start(self):
        """Default startup. Starts Simulators, Device Pool and MacroServer"""

        for s in self._getServsStartupOrder():
            s.start()

    def stop(self):
        """Default stop."""
                
        for s in self._getServsShutdownOrder():
            s.stop()
    
    def run(self):
        servs = self._getServsStartupOrder()
        self.o("Running ")
        count = 0
        while 1:
            count += 1
            if count % 10 == 0:
                self.o(":")
            for p in servs:
                p.run(step=True, timeout=0.1)
    
    def getDoc(self):
        return self._xmldoc
        
    def getRoot(self):
        return self._sarNode


if __name__ == "__main__":
    import getopt
    
    try:
        opts, pargs = getopt.getopt(sys.argv[1:], 'vl', ['simulation=','cleanup='])
    except Exception, e:
        print "ERROR:",str(e)
        print
        print __doc__
        sys.exit(3)
    
    if not len(pargs):
        print "ERROR: Please provide XML filename"
        print
        sys.exit(3)
 
    filename = pargs[0]
    
    simulation = "Best"
    cleanup = True
    just_output_and_exit = False
    activate_logging = False
    
    for opt,value in opts:
        if opt == '-l':
            activate_logging = True
        elif opt == '--simulation':
            simulation = value
        elif opt == '--cleanup':
            cleanup = value in ('yes', 'Yes', 'true', 'True', 'y', 'Y', '1')
        elif opt == '-v':
            just_output_and_exit = True
        else:
            print __doc__
            sys.exit(3)

    try:
        import to_sar
        sar_doc = to_sar.transform(filename)
    except Exception,e:
        print 'Sorry, but some problems found when trying to convert to SARDANA xml:'
        print str(e)
        
    sardana = Sardana(sar_doc, simulation=simulation, log=activate_logging, cleanup=cleanup)

    if just_output_and_exit:
        sardana.prepare()
        print etree.tostring(sardana.getRoot(), pretty_print=True)
        sys.exit(0)

    try:
        sardana.setUp()
        print "Ready!"
        sardana.run()
    except KeyboardInterrupt, e:
        print "User pressed Ctrl+C..."
    except Exception, e:
        traceback.print_exc()
        
    print "Shutting down!"
    sardana.tearDown()
    



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool library. It defines the base classes
for Pool object"""

__all__ = ["PoolBaseObject"]

__docformat__ = 'restructuredtext'

from sardana.sardanabase import SardanaBaseObject


class PoolBaseObject(SardanaBaseObject):
    """The Pool most abstract object."""
    
    def __init__(self, **kwargs):
        kwargs['manager'] = kwargs.pop('pool')
        SardanaBaseObject.__init__(self, **kwargs)
    
    def get_pool(self):
        """Return the :class:`sardana.pool.pool.Pool` which *owns* this pool
        object.
        
        :return: the pool which *owns* this pool object.
        :rtype: :class:`sardana.pool.pool.Pool`"""
        return self.get_manager()
    
    def serialize(self, *args, **kwargs):
        kwargs = SardanaBaseObject.serialize(self, *args, **kwargs)
        kwargs['pool'] = self.pool.name
        return kwargs
    
    pool = property(get_pool,
                    doc="reference to the :class:`sardana.pool.pool.Pool`")
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Sardana
## 
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This package provides the spock generic utilities"""

__all__ = ['page', 'arg_split', 'get_gui_mode', 'get_pylab_mode',
           'get_color_mode', 'get_ipapi',
           'get_editor', 'ask_yes_no', 'spock_input',
           'translate_version_str2int', 'get_ipython_version',
           'get_ipython_version_number', 'get_python_version',
           'get_python_version_number', 'get_ipython_dir',
           'get_ipython_profiles', 'get_spock_profiles', 
           'get_non_spock_profiles', 'get_spock_user_profile_module',
           'get_pytango_version', 'get_pytango_version_number',
           'get_server_for_device', 'get_macroserver_for_door',
           'get_device_from_user', 'get_tango_db', 'get_tango_host_from_user',
           'print_dev_from_class', 'from_name_to_tango', 'clean_up',
           'get_taurus_core_version', 'get_taurus_core_version_number',
           'check_requirements', 'get_door', 'get_macro_server', 'expose_magic',
           'unexpose_magic', 'expose_variable', 'expose_variables',
           'unexpose_variable',
           'create_spock_profile', 'check_for_upgrade', 'get_args',
           'init_console', 'init_magic', 'init_pre_spock', 'init_post_spock',
           'init_spock', 'start', 'mainloop', 'run',
           'load_ipython_extension', 'unload_ipython_extension', 'load_config',
           'MSG_FAILED','MSG_FAILED_WR', 'MSG_R', 'MSG_ERROR',
           'MSG_DONE', 'MSG_OK' ]

__docformat__ = 'restructuredtext'

import sys
import os
import socket
import imp

import IPython
import IPython.genutils
import PyTango
import PyTango.ipython

from taurus.core.taurushelper import Factory
from taurus.core.util.codecs import CodecFactory

from sardana.spock import exception
from sardana.spock import colors
from sardana.spock import release

arg_split = IPython.iplib.arg_split
page = IPython.genutils.page
TermColors = colors.TermColors


requirements = {
#     module     minimum  recommended 
    "IPython"     : ("0.10.0", "0.10.0"),
    "Python"      : ("2.6.0", "2.6.0"),
    "PyTango"     : ("7.1.2", "7.2.0"),
    "taurus.core" : ("2.0.0", "2.1.0")
}

ENV_NAME = "_E"

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# IPython utilities
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def get_gui_mode():
    if '-q4thread' in sys.argv:
        return 'qt'
    elif '-gthread' in sys.argv:
        return 'gtk'
    elif '-wthread' in sys.argv:
        return 'wx'
    return ''

def get_pylab_mode():
    return get_gui_mode()
    
def get_color_mode():
    return get_ipapi().options.colors
    
def get_ipapi():
    return IPython.ipapi.get()

def get_editor():
    return get_ipapi().options.editor

def ask_yes_no(prompt,default=None):
    """Asks a question and returns a boolean (y/n) answer.

    If default is given (one of 'y','n'), it is used if the user input is
    empty. Otherwise the question is repeated until an answer is given.

    An EOF is treated as the default answer.  If there is no default, an
    exception is raised to prevent infinite loops.

    Valid answers are: y/yes/n/no (match is not case sensitive)."""

    if default:
        prompt = '%s [%s]' % (prompt, default)
    return IPython.genutils.ask_yes_no(prompt, default)

def spock_input(prompt='',  ps2='... '):
    return IPython.genutils.raw_input_ext(prompt=prompt, ps2=ps2)   

def translate_version_str2int(version_str):
    """Translates a version string in format x[.y[.z[...]]] into a 000000 number"""
    import math
    parts = version_str.split('.')
    i, v, l = 0, 0, len(parts)
    if not l: return v
    while i<3:
        try:
            v += int(parts[i])*int(math.pow(10,(2-i)*2))
            l -= 1
            i += 1
        except ValueError:
            return v
        if not l: return v
    return v
    
    try:
        v += 10000*int(parts[0])
        l -= 1
    except ValueError:
        return v
    if not l: return v
    
    try:
        v += 100*int(parts[1])
        l -= 1
    except ValueError:
        return v
    if not l: return v

    try:
        v += int(parts[0])
        l -= 1
    except ValueError:
        return v
    if not l: return v

def get_ipython_version():
    """Returns the current IPython version"""
    v = None
    try:
        try:
            v = IPython.Release.version
        except Exception:
            try:
                v = IPython.release.version
            except Exception:
                pass
    except Exception:
        pass
    return v

def get_ipython_version_number():
    """Returns the current IPython version number"""
    ipyver_str = get_ipython_version()
    if ipyver_str is None: return None
    return translate_version_str2int(ipyver_str)

def get_python_version():
    return '.'.join(map(str,sys.version_info[:3]))

def get_python_version_number():
    pyver_str = get_python_version()
    return translate_version_str2int(pyver_str)

def get_ipython_dir():
    """Find the ipython local directory. Usually is <home>/.ipython"""
    if hasattr(PyTango.ipython, "get_ipython_dir"):
        return PyTango.ipython.get_ipython_dir()
    
    if hasattr(IPython.iplib, 'get_ipython_dir'):
        # Starting from ipython 0.9 they hadded this method
        return IPython.iplib.get_ipython_dir()
    
    # Try to find the profile in the current directory and then in the 
    # default IPython dir
    #userdir = os.path.realpath(os.path.curdir)
    home_dir = IPython.genutils.get_home_dir()
    
    if os.name == 'posix':
        ipdir = '.ipython'
    else:
        ipdir = '_ipython'
    ipdir = os.path.join(home_dir, ipdir)
    ipythondir = os.path.abspath( os.environ.get('IPYTHONDIR', ipdir) )
    return ipythondir

def get_ipython_profiles():
    """Helper function to find all ipython profiles"""
    if hasattr(PyTango.ipython, "get_ipython_profiles"):
        return PyTango.ipython.get_ipython_profiles()

    ret = []
    ipydir = get_ipython_dir()
    if os.path.isdir(ipydir):
        for i in os.listdir(ipydir):
            fullname = os.path.join(ipydir, i)
            if i.startswith("ipy_profile_") and i.endswith(".py"):
                if os.path.isfile(fullname):
                    ret.append(i[len("ipy_profile_"):i.rfind(".")])
    return ret

def get_spock_profiles(ipython_profiles=None):
    """Helper function to find all spock ipython profiles"""
    ret = []
    ipydir = get_ipython_dir()
    if not os.path.isdir(ipydir):
        return ret
    if ipython_profiles is None:
        ipython_profiles = get_ipython_profiles()
    ret = []
    for profile in ipython_profiles:
        profile_f = os.path.join(ipydir, "ipy_profile_%s.py" % profile)
        if not os.path.isfile(profile_f): continue
        try:
            for i, l in enumerate(file(profile_f)):
                if i > 10: break
                if l.find("spock_creation_version") >= 0:
                    ret.append(profile)
                    break
        except:
            pass
    return ret

def get_non_spock_profiles(ipython_profiles=None):
    """Helper function to find all non spock ipython profiles"""
    if ipython_profiles is None:
        ipython_profiles = get_ipython_profiles()
    ipython_profiles = set(ipython_profiles)
    spock_profiles = set(get_spock_profiles(ipython_profiles=ipython_profiles))
    return ipython_profiles.difference(spock_profiles)

def get_spock_user_profile_module(profile_name):
    return 'ipy_profile_%s' % profile_name
    
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# PyTango utilities
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def get_pytango_version():
    try:
        import PyTango
    except:
        return None
    try:
        return PyTango.Release.version
    except:
        return '0.0.0'

def get_pytango_version_number():
    tgver_str = get_pytango_version()
    if tgver_str is None: return None
    return translate_version_str2int(tgver_str)
    
def get_server_for_device(device_name):
    db = get_tango_db()
    device_name = device_name.lower()
    server_list = db.get_server_list()
    for server in server_list:
        for dev in db.get_device_class_list(server)[::2]:
            if dev.lower() == device_name:
                return server
    return None
    
def get_macroserver_for_door(door_name):
    """Returns the MacroServer device name in the same DeviceServer as the
    given door device"""
    full_door_name, door_name, door_alias = from_name_to_tango(door_name)
    db = get_tango_db()
    door_name = door_name.lower()
    server_list = list(db.get_server_list('MacroServer/*'))
    server_list += list(db.get_server_list('Sardana/*'))
    server_devs = None
    for server in server_list:
        server_devs = db.get_device_class_list(server)
        devs, klasses = server_devs[0::2], server_devs[1::2]
        for dev in devs:
            if dev.lower() == door_name:
                for i, klass in enumerate(klasses):
                    if klass == 'MacroServer':
                        return "%s:%s/%s" % (db.get_db_host(), db.get_db_port(), devs[i])
    else:
        return None

def get_device_from_user(expected_class, dft = None):
    """Gets a device of the given device class from user input"""
    dft = print_dev_from_class(expected_class, dft)
    prompt = "%s name from the list" % expected_class
    if not dft is None:
        prompt += "[%s]" % dft
    prompt += "? "
    from_user = raw_input(prompt).strip() or dft
    
    name = ''
    try:
        full_name, name, alias = from_name_to_tango(from_user)
    except:
        print "Warning: the given %s does not exist" % expected_class
        return name 
    
    try:
        db = get_tango_db()
        cl_name = db.get_class_for_device(name)
        class_correct = cl_name == expected_class
        if not class_correct:
            print "Warning: the given name is not a %s (it is a %s)"%(expected_class,cl_name)
    except Exception as e:
        print "Warning: unable to confirm if '%s' is valid" % name
        print str(e)
    return full_name

def get_tango_db():
    tg_host = PyTango.ApiUtil.get_env_var("TANGO_HOST")

    db = None
    if tg_host is None:
        host,port = get_tango_host_from_user()
        tg_host = "%s:%d" % (host,port)
        os.environ["TANGO_HOST"] = tg_host
        db = PyTango.Database()
    else:
        try:
            db = PyTango.Database()
        except:
            # tg host is not valid. Find a valid one
            host,port = get_tango_host_from_user()
            tg_host = "%s:%d" % (host,port)
            os.environ["TANGO_HOST"] = tg_host
            
            db = PyTango.Database()
    return db

def get_tango_host_from_user():

    while True: 
        prompt = "Please enter a valid tango host (<host>:<port>): "
        from_user = raw_input(prompt).strip()
    
        try:
            host, port = from_user.split(':')
            try:
                port = int(port)
                try:
                    socket.gethostbyname(host)
                    try:
                        db = PyTango.Database(host,port)
                        return db.get_db_host(), db.get_db_port()
                    except:
                        exp = "No tango database found at %s:%d" % (host,port)
                except:
                    exp = "Invalid host name %s" % host
            except:
                exp = "Port must be a number > 0"
        except:
            exp = "Invalid tango host. Must be in format <host>:<port>"
        exp = "Invalid tango host. %s " % exp
        print exp

def print_dev_from_class(classname, dft = None):

    db = get_tango_db()
    pytg_ver = get_pytango_version_number()
    if pytg_ver >= 030004:
        server_wildcard = '*'
        try:
            exp_dev_list = db.get_device_exported_for_class(classname)
        except Exception:
            exp_dev_list = []
    else:
        server_wildcard = '%'
        exp_dev_list = []
    
    res = None
    dev_list = db.get_device_name(server_wildcard,classname)
    tg_host = "%s:%s" % (db.get_db_host(),db.get_db_port())
    print "Available",classname,"devices from",tg_host,":"
    for dev in dev_list:
        full_name, name, alias = from_name_to_tango(dev)
        out = alias or name
        if alias: out += ' (a.k.a. %s)' % name
        out = "%-25s" % out 
        if dev in exp_dev_list:
            out += " (running)"
        print out
        if dft:
            if dft.lower() == name.lower(): res = name
            elif not alias is None and dft.lower() == alias.lower(): res = alias 
    
    return res
    
def from_name_to_tango(name):
    
    db = get_tango_db()
    
    alias = None
    
    c = name.count('/')
    # if the db prefix is there, remove it first
    if c == 3 or c == 1:
        name = name[name.index("/")+1:]
    
    elems = name.split('/')
    l = len(elems)
    
    if l == 3:
        try:
            alias = db.get_alias(name)
            if alias.lower() == 'nada':
                alias = None
        except:
            alias = None
    elif l == 1:
        alias = name
        name = db.get_device_alias(alias)
    else:
        raise Exception("Invalid device name '%s'" % name)
    
    full_name = "%s:%s/%s" % (db.get_db_host(), db.get_db_port(), name)
    return full_name, name, alias

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# taurus utilities
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def clean_up():
    taurus.Manager().cleanUp()

def get_taurus_core_version():
    try:
        import taurus
        return taurus.core.release.version
    except:
        return '0.0.0'
        
def get_taurus_core_version_number():
    tgver_str = get_taurus_core_version()
    if tgver_str is None: return None
    return translate_version_str2int(tgver_str)

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Requirements checking
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def check_requirements():
    r = requirements
    minPyTango, recPyTango = map(translate_version_str2int, r["PyTango"])
    minIPython, recIPython = map(translate_version_str2int, r["IPython"])
    minPython, recPython = map(translate_version_str2int, r["Python"])
    minTaurusCore, recTaurusCore = map(translate_version_str2int, r["taurus.core"])
    
    currPython = get_python_version_number()
    currIPython = get_ipython_version_number()
    currPyTango = get_pytango_version_number()
    currTaurusCore = get_taurus_core_version_number()
    
    errMsg = ""
    warnMsg = ""
    
    errPython, errIPython, errPyTango, errTaurusCore = False, False, False, False
    if currPython is None:
        errMsg += "Spock needs Python version >= %s. No python installation found\n" % requirements["Python"][0]
        errPython = True
    elif currPython < minPython:
        errMsg += "Spock needs Python version >= %s. Current version is %s\n" % (requirements["Python"][0], get_python_version())
        errPython = True

    if currIPython is None:
        errMsg += "Spock needs IPython version >= %s. No IPython installation found\n" % requirements["IPython"][0]
        errIPython = True
    elif currIPython < minIPython:
        errMsg += "Spock needs IPython version >= %s. Current version is %s\n" % (requirements["IPython"][0], get_ipython_version())
        errIPython = True

    if currPyTango is None:
        errMsg += "Spock needs PyTango version >= %s. No PyTango installation found\n" % requirements["IPython"][0]
        errPyTango = True
    elif currPyTango < minPyTango:
        errMsg += "Spock needs PyTango version >= %s. " % requirements["PyTango"][0]
        if currPyTango > 0:
            errMsg += "Current version is %s\n" % get_pytango_version()
        else:
            errMsg += "Current version is unknown (most surely too old)\n"
        errPyTango = True

    if currTaurusCore is None:
        errMsg += "Spock needs taurus.core version >= %s. No taurus.core installation found\n" % requirements["taurus.core"][0]
        errTaurusCore = True
    elif currTaurusCore < minTaurusCore:
        errMsg += "Spock needs taurus.core version >= %s. " % requirements["taurus.core"][0]
        if currTaurusCore > 0:
            errMsg += "Current version is %s\n" % get_taurus_core_version()
        else:
            errMsg += "Current version is unknown (most surely too old)\n"
        errTaurusCore = True
        
    # Warnings
    if not errPython and currPython < recPython:
        warnMsg += "Spock recommends Python version >= %s. Current version is %s\n" % (requirements["Python"][1], get_python_version())

    if not errIPython and currIPython < recIPython:
        warnMsg += "Spock recommends IPython version >= %s. Current version is %s\n" % (requirements["IPython"][1], get_ipython_version())

    if not errPyTango and currPyTango < recPyTango:
        warnMsg += "Spock recommends PyTango version >= %s. Current version is %s\n" % (requirements["PyTango"][1], get_pytango_version())
            
    if not errTaurusCore and currTaurusCore < recTaurusCore:
        warnMsg += "Spock recommends taurus.core version >= %s. Current version is %s\n" % (requirements["taurus.core"][1], get_taurus_core_version())

    if errMsg:
        errMsg += warnMsg
        raise exception.SpockMissingRequirement, errMsg

    if warnMsg: 
        raise exception.SpockMissingRecommended, warnMsg
    
    return True

def _get_dev(dev_type):
    ip = get_ipapi()
    ret = ip.user_ns.get("_" + dev_type)
    if ret is not None:
        return ret
    
    dev_obj_name = '%s_NAME' % dev_type
    dev_name = ip.user_ns[dev_obj_name]
    factory = Factory()
    dev_obj = factory.getDevice(dev_name)
    ip.user_ns[dev_type] = PyTango.DeviceProxy(dev_name) 
    ip.user_ns["_" + dev_type] = dev_obj
    setattr(ip, '_%s' % dev_type, dev_obj)
    return dev_obj

def get_door():
    return _get_dev('DOOR')

def get_macro_server():
    return _get_dev('MACRO_SERVER')

def _macro_completer(self, event):
    """Method called by the IPython autocompleter. It will determine possible
       values for macro arguments.
    """
    ms = get_macro_server()

    macro_name = event.command.lstrip('%')
    
    # calculate parameter index
    param_idx = len(event.line.split()) - 1
    if not event.line.endswith(' '): param_idx -= 1
    # get macro info
    info = ms.getMacroInfoObj(macro_name)
    # if macro doesn't have parameters return
    if param_idx < 0 or not info.hasParams() : return
    # get the parameter info
    possible_params = info.getPossibleParams(param_idx)
    # return the existing elements for the given parameter type
    if possible_params:
        res = []
        for param in possible_params:
            res.extend(ms.getElementNamesWithInterface(param['type']))
        return res
        
def expose_magic(name, fn, completer_func=_macro_completer):
    ip = get_ipapi()
    ip.expose_magic(name, fn)
    
    if completer_func is None:
        return
    
    # enable macro param completion
    ip.set_hook('complete_command', completer_func, str_key = name)

    # register also when the command as is typed with the magic prefix '%'
    name = str('%') + name
    ip.set_hook('complete_command', completer_func, str_key = name)
    
def unexpose_magic(name):
    ip = get_ipapi()
    mg = 'magic_%s' % name
    delattr(ip.IP, mg)

def expose_variable(name, value):
    get_ipapi().to_user_ns({ name : value })

def unexpose_variable(name):
    user_ns = get_ipapi().user_ns
    del user_ns[name]
    
def expose_variables(d):
    get_ipapi().to_user_ns(d)

#def _expose_device(name):
#    ip.to_user_ns({ name : PyTango.DeviceProxy(name) })
    
#def expose_device(name):
#    ip = get_ipapi()
#    ip.magic("bg _expose_device(%s)" % name)

def create_spock_profile(userdir, dft_profile, profile, door_name=None):
    """Create a profile file from a profile template file """

    src_data = """\
\"\"\"Settings for Spock session\"\"\"

#
# Please do not delete the next lines has they are used to check the version 
# number for possible upgrades
# spock_creation_version = {version}
# door_name = {door_name}
#

import IPython
from sardana.spock.genutils import init_spock

def main():
    ip = IPython.ipapi.get()
    init_spock(ip, '{macroserver_name}', '{door_name}')

main()
"""
    
    #
    # Discover door name
    #
    if door_name is None:
        door_name = get_device_from_user("Door", profile)
    else:
        full_door_name, door_name, door_alias = from_name_to_tango(door_name)
        door_name = full_door_name
    
    #
    # Discover macro server name
    #
    ms_name = get_macroserver_for_door(door_name)
    
    dest_data = src_data.format(version=release.version,
                                macroserver_name=ms_name,
                                door_name=door_name)
    
    f_name = '%s.py' % get_spock_user_profile_module(profile)
    
    dest_name = os.path.join(userdir, f_name)
    
    sys.stdout.write('Storing %s in %s... ' % (f_name, userdir))
    sys.stdout.flush()
    res = MSG_FAILED
    try:
        dest = open(dest_name,"w")
        dest.write(dest_data)
        dest.flush()
        dest.close()
        res = MSG_DONE
    finally:
        sys.stdout.write(res + '\n')
        sys.stdout.flush()

def check_for_upgrade(ipy_profile_file, ipythondir, session, profile):
    # Check if the current profile is up to date with the spock version
    spock_profile_ver_str = '0.0.0'
    door_name = None
    
    # search for version and door inside the ipy_profile file
    for i, line in enumerate(ipy_profile_file):
        if i > 20 : break; # give up after 20 lines
        if line.startswith('# spock_creation_version = '):
            spock_profile_ver_str = line[line.index('=')+1:].strip()
        if line.startswith('# door_name = '):
            door_name = line[line.index('=')+1:].strip()
    
    # convert version from string to numbers
    spocklib_ver = translate_version_str2int(release.version)
    spock_profile_ver = translate_version_str2int(spock_profile_ver_str)
    
    if spocklib_ver == spock_profile_ver:
        return
    if spocklib_ver < spock_profile_ver:
        print '%sYour spock profile (%s) is newer than your spock version ' \
              '(%s)!' % (TermColors.Brown, spock_profile_ver_str, release.version)
        print 'Please upgrade spock or delete the current profile %s' % TermColors.Normal
        sys.exit(1)
        
    # there was no version track of spock profiles since spock 0.2.0 so change
    # the message
    if spock_profile_ver_str == '0.0.0':
        spock_profile_ver_str = '<= 0.2.0'
    msg = 'Your current spock door extension profile has been created with spock %s.\n' \
          'Your current spock door extension version is %s, therefore a profile upgrade is needed.\n' \
          % (spock_profile_ver_str, release.version)
    print msg
    prompt = 'Do you wish to upgrade now (warn: this will shutdown the current spock session) ([y]/n)? '
    r = raw_input(prompt) or 'y'
    if r.lower() == 'y':
        create_spock_profile(ipythondir, session, profile, door_name)
        sys.exit(0)

def get_args(argv):
    
    script_name = argv[0]
    script_dir, session = os.path.split(script_name)
    script_name = os.path.realpath(script_name)
    #script_dir = os.path.dirname(script_name)
    
    macro_server = None
    door = None
    
    # Define the profile file
    profile = "spockdoor"
    try:
        profile_idx = argv.index('-p') + 1
        if len(argv) > profile_idx:
            profile = argv[profile_idx]
    except:
        pass

    profile_modulename = get_spock_user_profile_module(profile)
    
    # Try to find the profile in the current directory and then in the 
    # default IPython dir
    #userdir = os.path.realpath(os.path.curdir)
    ipythondir = get_ipython_dir()
    
    if not os.path.isdir(ipythondir):
        # Platform-dependent suffix.
        if os.name == 'posix':
            rc_suffix = ''
        else:
            rc_suffix = '.ini'
        IPython.iplib.user_setup(ipythondir, rc_suffix, mode='install',
                                 interactive=False)
    
    try:
        f, name, t = imp.find_module(profile_modulename, [ipythondir])
        check_for_upgrade(f, ipythondir, session, profile)
    except ImportError:
        # Create a new profile
        r = ''
        while not r in ['y','n']:
            prompt = 'Profile \'%s\' does not exist. Do you want to create '\
                     'one now ([y]/n)? ' % profile
            r = raw_input(prompt) or 'y'
        if r.lower() == 'y':
            create_spock_profile(ipythondir, session, profile)
        else:
            sys.stdout.write('No spock door extension profile was created. Starting normal spock...\n')
            sys.stdout.flush()
            profile = ''
    
    # inform the shell of the profile it should use
    if not '-p' in argv and profile:
        argv.append('-p')
        argv.append(profile)
    
    user_ns = { 'MACRO_SERVER_NAME' : macro_server,
                'DOOR_NAME'         : door,
                'PROFILE'           : profile }
    
    return user_ns

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Useful constants
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

MSG_G = '[%s%%s%s]' % (TermColors.Green, TermColors.Normal)
MSG_R = '[%s%%s%s]' % (TermColors.Red, TermColors.Normal)
MSG_FAILED = MSG_R % 'FAILED'
MSG_FAILED_WR = MSG_R % 'FAILED: %s'
MSG_ERROR  = MSG_R % 'ERROR'
MSG_DONE   = MSG_G % 'DONE'
MSG_OK     = MSG_G % 'OK'

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# initialization methods
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def init_console(ip):
    # Handy tab-completers for %cd, %run, import etc.
    # Try commenting this out if you have completion problems/slowness
    import ipy_stock_completers

    spockver = release.version
    pyver = get_python_version()
    ipyver = get_ipython_version()
    pytangover = get_pytango_version()
    tauruscorever = get_taurus_core_version()
    
    TermColors = IPython.ColorANSI.TermColors
    
    d = { "version" : spockver,
          "pyver" : pyver,
          "ipyver" : ipyver,
          "pytangover" : pytangover,
          "taurusver" : tauruscorever,
          "profile" : ip.user_ns["PROFILE"],
          "door" : ip.user_ns["DOOR_ALIAS"] }
          
    d.update(TermColors.__dict__)

    # IPython options
    o = ip.options
    o.autocall = 1
    o.autoedit_syntax = 0
    o.autoindent = 1
    o.automagic = 1
    o.cache_size = 1000
    o.colors = 'Linux'
    o.color_info = 1
    o.confirm_exit = 0
    o.deep_reload = 0
    #o.editor = 'gedit'
    o.log = 0
    o.logfile = ''
    o.messages = 1
    o.pdb = 0
    o.pprint = 1
    o.quick = 0
    o.readline = 1
    o.screen_length = 0
    o.separate_in = '\n'
    o.separate_out = '\n'
    o.separate_out2 = ''
    o.nosep = 0
    o.wildcards_case_sensitive = 0
    o.object_info_string_level = 0
    o.xmode = 'Context'
    o.multi_line_specials = 1
    o.system_header = 'IPython system call: '
    o.system_verbose = 1
    o.wxversion = '0'
    o.colors = "GreenTango"
    o.prompt_in1 = "$DOOR_ALIAS$DOOR_STATE [\\#]: "
    o.prompt_out = "Result [\\#]: "
    o.readline_parse_and_bind.append('tab: complete')
    #o.readline_parse_and_bind.append('tab: menu-complete')
    o.readline_parse_and_bind.append('"\C-l": possible-completions')
    o.readline_parse_and_bind.append('set show-all-if-ambiguous on')
    o.readline_parse_and_bind.append('"\C-o": tab-insert')
    o.readline_parse_and_bind.append('"\M-i": "    "')
    o.readline_parse_and_bind.append('"\M-o": "\d\d\d\d"')
    o.readline_parse_and_bind.append('"\M-I": "\d\d\d\d"')
    o.readline_parse_and_bind.append('"\C-r": reverse-search-history')
    o.readline_parse_and_bind.append('"\C-s": forward-search-history')
    o.readline_parse_and_bind.append('"\C-p": history-search-backward')
    o.readline_parse_and_bind.append('"\C-n": history-search-forward')
    o.readline_parse_and_bind.append('"\e[A": history-search-backward')
    o.readline_parse_and_bind.append('"\e[B": history-search-forward')
    o.readline_parse_and_bind.append('"\C-k": kill-line')
    o.readline_parse_and_bind.append('"\C-u": unix-line-discard')
    o.readline_remove_delims = '-/~'
    o.readline_merge_completions = 1
    o.readline_omit__names = 0

    banner = """\
%(Purple)sSpock %(version)s%(Normal)s -- An interactive laboratory application.

help      -> Spock's help system.
object?   -> Details about 'object'. ?object also works, ?? prints more.
"""
    banner = banner % d
    banner = banner.format(**d)

    o.banner = banner

def init_magic(ip):
    import sardana.spock.magic
    magic = sardana.spock.magic
    expose_magic('debug', magic.debug, magic.debug_completer)
    expose_magic('www', magic.www, None)
    expose_magic('post_mortem', magic.post_mortem, None)
    expose_magic('spsplot', magic.spsplot, None)
    expose_magic('macrodata', magic.macrodata, None)
    expose_magic('edmac', magic.edmac, None)
    expose_magic('showscan', magic.showscan, None)
    expose_magic('expconf', magic.expconf, None)
    ip.set_hook('late_startup_hook', magic.spock_late_startup_hook)
    ip.set_hook('pre_prompt_hook', magic.spock_pre_prompt_hook)

def init_pre_spock(ip, macro_server, door):
    so = IPython.ipstruct.Struct()
    full_door_tg_name, door_tg_name, door_tg_alias = from_name_to_tango(door)
    #macro_server = get_ms_for_door(door_tg_name)
    full_ms_tg_name, ms_tg_name, ms_tg_alias = from_name_to_tango(macro_server)
    ip.user_ns['MACRO_SERVER_NAME'] = full_ms_tg_name
    ip.user_ns['MACRO_SERVER_ALIAS'] = ms_tg_alias or ms_tg_name
    ip.user_ns['DOOR_NAME'] = full_door_tg_name
    ip.user_ns['DOOR_ALIAS'] = door_tg_alias or door_tg_name
    ip.user_ns['DOOR_STATE'] = ""
    ip.user_ns['spock_options'] = so
    
    if ip.IP.alias_table.has_key('mv'):
        del ip.IP.alias_table['mv']
    
    v = release.version
    alias = ip.user_ns['DOOR_ALIAS']
    profile = ip.user_ns['PROFILE']
    
    so.spock_banner = """\
{Blue}Spock's sardana extension %s loaded with profile: %s (linked to door '%s'){Normal}
""" % (v, profile, alias)

    # the CodecFactory is not thread safe. There are two attributes who will
    # request for it in the first event at startup in different threads
    # therefore this small hack: make sure CodecFactory is initialized.
    CodecFactory()
    
    factory = Factory()

    import sardana.spock.spockms
    macroserver = sardana.spock.spockms

    factory.registerDeviceClass('MacroServer', macroserver.SpockMacroServer)
    
    mode = get_gui_mode()
    if mode == 'qt':
        factory.registerDeviceClass('Door', macroserver.QSpockDoor)
    else:
        factory.registerDeviceClass('Door', macroserver.SpockDoor)
    
    door = get_door()
    macro_server = get_macro_server()
    
    # Initialize the environment
    expose_variable(ENV_NAME, macro_server.getEnvironment())


def init_post_spock(ip):
    init_console(ip)
    init_magic(ip)
    
def init_spock(ip, macro_server, door):
    init_pre_spock(ip, macro_server, door)
    PyTango.ipython.init_ipython(ip)
    init_post_spock(ip)


def start(user_ns=None):
    if '-pylab' not in sys.argv: sys.argv.insert(1, '-pylab')
    if '-q4thread' not in sys.argv: sys.argv.insert(1, '-q4thread')

    # Make sure the log level is changed to warning
    from taurus.core.taurushelper import setLogLevel, Warning
    CodecFactory()
    setLogLevel(Warning)

    try:
        check_requirements()
    except exception.SpockMissingRequirement, requirement:
        print str(requirement)
        sys.exit(-1)
    except exception.SpockMissingRecommended, recommended:
        print str(recommended)
    
    user_ns = user_ns or {}
    try:
        user_ns.update(get_args(sys.argv))
    except exception.SpockException, e:
        print e.message
        print 'Starting normal IPython console'
    except KeyboardInterrupt:
        print "\nUser pressed Ctrl+C. Exiting..."
        sys.exit()
    except Exception, e:
        print 'spock exited with an unmanaged exception: %s' % str(e)
        sys.exit(-2)
        
    return IPython.Shell.start(user_ns=user_ns)

def mainloop(shell=None, user_ns=None):
    if shell is None:
        shell = start(user_ns)
    shell.mainloop()

def run(user_ns=None):
    # initialize input handler as soon as possible
    import sardana.spock.inputhandler
    input_handler = sardana.spock.inputhandler.InputHandler()

    try:
        mainloop(user_ns=user_ns)
    finally:
        try:
            clean_up()
        except Exception:
            pass

# for compatibility reasons with new IPython API (>=0.11) we add the following
# empty methods
            
def load_ipython_extension(ipython):
    pass

def unload_ipython_extension(ipython):
    pass
    
def load_config(config):
    pass
